///////////////////////////////////////////////////////////////////////////////
//
/// \file       alone_decoder.c
/// \brief      Decoder for LZMA_Alone files
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "alone_common.h"
#include "lzma_decoder.h"


static lzma_ret
alone_decode(lzma_coder *coder,
		lzma_allocator *allocator lzma_attribute((unused)),
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size,
		lzma_action action)
{
	while (*out_pos < out_size
			&& (coder->sequence == SEQ_CODE || *in_pos < in_size))
	switch (coder->sequence) {
	case SEQ_PROPERTIES:
		if (lzma_lzma_decode_properties(
				&coder->options, in[*in_pos])) {
			coder->sequence = SEQ_UNINITIALIZED;
			return LZMA_DATA_ERROR;
		}

		coder->sequence = SEQ_DICTIONARY_SIZE;
		++*in_pos;
		break;

	case SEQ_DICTIONARY_SIZE:
		coder->options.dictionary_size
				|= (size_t)(in[*in_pos]) << (coder->pos * 8);

		if (++coder->pos == 4) {
			// A hack to ditch tons of false positives: We allow
			// only dictionary sizes that are a power of two.
			// LZMA_Alone didn't create other kinds of files,
			// although it's not impossible that files with
			// other dictionary sizes exist. Well, if someone
			// complains, this will be reconsidered.
			size_t count = 0;
			for (size_t i = 0; i < 32; ++i)
				if (coder->options.dictionary_size
						& (UINT32_C(1) << i))
					++count;

			if (count > 1)
				return LZMA_DATA_ERROR;

			coder->pos = 0;
			coder->sequence = SEQ_UNCOMPRESSED_SIZE;
		}

		++*in_pos;
		break;

	case SEQ_UNCOMPRESSED_SIZE:
		coder->uncompressed_size
				|= (uint64_t)(in[*in_pos]) << (coder->pos * 8);

		if (++coder->pos == 8) {
			// Another hack to ditch false positives: Assume that
			// if the uncompressed size is known, it must be less
			// than 256 GiB. Again, if someone complains, this
			// will be reconsidered.
			if (coder->uncompressed_size != UINT64_MAX
					&& coder->uncompressed_size >=
						(UINT64_C(1) << 38))
				return LZMA_DATA_ERROR;

			coder->pos = 0;
			coder->sequence = SEQ_CODER_INIT;
		}

		++*in_pos;
		break;

	case SEQ_CODER_INIT: {
		// Two is enough because there won't be implicit filters.
		lzma_filter_info filters[2] = {
			{
				.init = &lzma_lzma_decoder_init,
				.options = &coder->options,
				.uncompressed_size = coder->uncompressed_size,
			}, {
				.init = NULL,
			}
		};

		const lzma_ret ret = lzma_next_coder_init(&coder->next,
				allocator, filters);

		if (ret != LZMA_OK) {
			coder->sequence = SEQ_UNINITIALIZED;
			return ret;
		}

		coder->sequence = SEQ_CODE;
	}

	// Fall through

	case SEQ_CODE: {
		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

		switch (ret) {
		case LZMA_OK:
			assert(*in_pos == in_size || *out_pos == out_size);
			return LZMA_OK;

		case LZMA_STREAM_END:
			coder->sequence = SEQ_END;
			return LZMA_STREAM_END;

		default:
			assert(ret == LZMA_DATA_ERROR);
			coder->sequence = SEQ_UNINITIALIZED;
			return ret;
		}

		break;
	}

	case SEQ_END:
		return LZMA_STREAM_END;

	case SEQ_UNINITIALIZED:
	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
alone_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_alone_decoder_init(lzma_next_coder *next, lzma_allocator *allocator)
{
	if (next->init != &lzma_alone_decoder_init) {
		// Get rid of the previous coder.
		lzma_next_coder_end(next, allocator);

		// Allocate data structure for LZMA_Alone coder.
		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		// Mark that we don't have any LZMA coder initialized yet.
		next->coder->next = LZMA_NEXT_CODER_INIT;
	}

	// Initialize the LZMA_Alone coder variables.
	next->coder->sequence = SEQ_PROPERTIES;
	next->coder->pos = 0;
	next->coder->options.dictionary_size = 0;
	next->coder->uncompressed_size = 0;

	// Initialization successful, set the function pointers.
	next->init = &lzma_alone_decoder_init;
	next->code = &alone_decode;
	next->end = &alone_decoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_alone_decoder(lzma_stream *strm)
{
	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FLUSH] = true;

	ret = lzma_alone_decoder_init(&strm->internal->next, strm->allocator);
	if (ret != LZMA_OK)
		lzma_end(strm);

	return ret;
}
