///////////////////////////////////////////////////////////////////////////////
//
/// \file       block_encoder.c
/// \brief      Encodes .lzma Blocks
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "block_encoder.h"
#include "raw_encoder.h"
#include "check.h"


struct lzma_coder_s {
	/// The filters in the chain; initialized with lzma_raw_decoder_init().
	lzma_next_coder next;

	enum {
		SEQ_HEADER,
		SEQ_CODE,
		SEQ_CHECK,
	} sequence;

	/// Position in .header and .check.
	size_t pos;

	/// Block Header in encoded form. It is allocated and encoded when
	/// initializing the Block encoder.
	uint8_t *header;
	size_t header_size;

	/// Check of the uncompressed data
	lzma_check check;
	lzma_check_type check_type;

	// Size variables updated when encoding.
	uint64_t compressed_size;
	uint64_t uncompressed_size;

	// Reference values given to us when initializing. These are
	// also stored in the Block Header.
	uint64_t compressed_size_in_header;
	uint64_t uncompressed_size_in_header;
};


static lzma_ret
block_encode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size
			&& (*in_pos < in_size || action == LZMA_FINISH))
	switch (coder->sequence) {
	case SEQ_HEADER:
		bufcpy(coder->header, &coder->pos, coder->header_size,
				out, out_pos, out_size);

		if (coder->pos == coder->header_size) {
			lzma_free(coder->header, allocator);
			coder->header = NULL;

			coder->pos = 0;
			coder->sequence = SEQ_CODE;
		}

		break;

	case SEQ_CODE: {
		// If Uncompressed Size was given, it must match once
		// action == LZMA_FINISH.
		if (action == LZMA_FINISH
				&& coder->uncompressed_size_in_header
					!= UINT64_MAX
				&& coder->uncompressed_size_in_header
					!= coder->uncompressed_size
						+ (in_size - *in_pos))
			return LZMA_DATA_ERROR;

		// Store the old positions to be used when updating
		// coder->compressed_size, coder->uncompressed_size, and
		// coder->data_check.
		const size_t in_start = *in_pos;
		const size_t out_start = *out_pos;

		// Encode
		const lzma_ret ret = coder->next.code(coder->next.coder,
				allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

		// Update coder->check. We ignore the return value because
		// we checked that check_type is valid when coder->check
		// was initialized.
		if (in_start < *in_pos)
			(void)lzma_check_update(
					&coder->check, coder->check_type,
					in + in_start, *in_pos - in_start);

		// Update coder->uncompressed_size and coder->compressed_size.
		coder->uncompressed_size += *in_pos - in_start;
		coder->compressed_size += *out_pos - out_start;

		// Verify that the sizes stay in valid range.
		if (coder->compressed_size > coder->compressed_size_in_header
				|| coder->uncompressed_size
					> coder->uncompressed_size_in_header)
			return LZMA_DATA_ERROR;

		switch (ret) {
		case LZMA_OK:
			assert(*in_pos == in_size || *out_pos == out_size);
			return LZMA_OK;

		case LZMA_STREAM_END:
			// If Compressed Size was given, it must match now.
			if (coder->compressed_size_in_header != UINT64_MAX
					&& coder->compressed_size
					!= coder->compressed_size_in_header)
				return LZMA_DATA_ERROR;

			// If Uncompressed Size was given, it must match now.
			if (coder->uncompressed_size_in_header != UINT64_MAX
					&& coder->uncompressed_size
					!= coder->uncompressed_size_in_header)
				return LZMA_DATA_ERROR;

			// If no Check is used, we are done now.
			if (coder->check_type == LZMA_CHECK_NONE)
				return LZMA_STREAM_END;

			lzma_check_finish(&coder->check, coder->check_type);
			coder->sequence = SEQ_CHECK;
			break;

		default:
			return ret;
		}

		break;
	}

	case SEQ_CHECK:
		assert(coder->check_type > LZMA_CHECK_NONE);
		assert(coder->check_type <= 7);
		assert(lzma_check_sizes[coder->check_type] > 0);

		switch (coder->check_type) {
		case LZMA_CHECK_CRC32:
			out[*out_pos] = coder->check.crc32 >> (coder->pos * 8);
			break;

		case LZMA_CHECK_CRC64:
			out[*out_pos] = coder->check.crc64 >> (coder->pos * 8);
			break;

		case LZMA_CHECK_SHA256:
			out[*out_pos] = coder->check.sha256.buffer[coder->pos];
			break;

		default:
			assert(0);
			return LZMA_STREAM_ERROR;
		}

		++*out_pos;

		if (++coder->pos == lzma_check_sizes[coder->check_type])
			return LZMA_STREAM_END;

		break;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
block_encoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_free(coder->header, allocator);
	lzma_next_coder_end(&coder->next, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_block_encoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_options_block *options)
{
	// Allocate and initialize *next->coder if needed.
	if (next->init != &lzma_block_encoder_init) {
		// Free the old coder.
		lzma_next_coder_end(next, allocator);

		// Allocate a new Block encoder structure.
		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		// Mark that we haven't initialized any filters yet.
		next->coder->next = LZMA_NEXT_CODER_INIT;

		// Mark that we haven't allocated any memory for header yet.
		next->coder->header = NULL;
	}

	lzma_ret ret;

	// Validate some options.
	if (options->check > LZMA_CHECK_ID_MAX
			|| !lzma_vli_is_valid(options->uncompressed_size)
			|| !lzma_vli_is_valid(options->compressed_size)) {
		ret = LZMA_STREAM_ERROR;
		goto error;
	}

	// Reset some of the variables. These are initialized in
	// encode_header(): header, header_size, padding
	next->coder->sequence = SEQ_HEADER;
	next->coder->pos = 0;
	next->coder->check_type = options->check;
	next->coder->compressed_size = 0;
	next->coder->uncompressed_size = 0;
	next->coder->compressed_size_in_header = options->compressed_size;
	next->coder->uncompressed_size_in_header = options->uncompressed_size;

	// Initialize the check.
	if (lzma_check_init(&next->coder->check, next->coder->check_type)) {
		ret = LZMA_UNSUPPORTED_CHECK;
		goto error;
	}

	// Initialize the requested filters.
	ret = lzma_raw_encoder_init(&next->coder->next, allocator,
			options->filters, options->has_eopm ? UINT64_MAX
				: options->uncompressed_size, true);
	if (ret != LZMA_OK)
		goto error;

	// Encode *options to coder->filters. Don't include Metadata filter
	// if it has been used.
	lzma_free(next->coder->header, allocator);
	ret = lzma_block_header_encode(&next->coder->header,
			&next->coder->header_size, allocator, options);
	if (ret != LZMA_OK)
		goto error;

	// All successfully initialized, set the function pointers.
	next->init = &lzma_block_encoder_init;
	next->code = &block_encode;
	next->end = &block_encoder_end;

	return LZMA_OK;

error:
	block_encoder_end(next->coder, allocator);
	return ret;
}


extern LZMA_API lzma_ret
lzma_block_encoder(lzma_stream *strm, const lzma_options_block *options)
{
	if (options == NULL)
		return LZMA_STREAM_ERROR;

	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;
	strm->internal->supported_actions[LZMA_FINISH] = true;

	ret = lzma_block_encoder_init(
			&strm->internal->next, strm->allocator, options);

	if (ret != LZMA_OK)
		(void)lzma_end(strm);

	return ret;
}
