///////////////////////////////////////////////////////////////////////////////
//
/// \file       filter_flags_decoder.c
/// \brief      Decodes a Filter Flags field
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "filter_flags_common.h"
#include "lzma_decoder.h"


struct lzma_coder_s {
	lzma_options_filter *options;

	enum {
		SEQ_MISC,
		SEQ_ID,
		SEQ_SIZE,
		SEQ_PROPERTIES,
	} sequence;

	/// \brief      Position in variable-length integers
	size_t pos;

	/// \brief      Size of Filter Properties
	uint64_t properties_size;
};


/// \brief      Decodes LZMA Properties and Dictionary Flags (two bytes)
static lzma_ret
properties_lzma(lzma_options_lzma *options, size_t *prop_pos, size_t prop_size,
		const uint8_t *in, size_t *in_pos, size_t in_size)
{
	if (prop_size != 2)
		return LZMA_HEADER_ERROR;

	assert(*prop_pos < 2);

	while (*in_pos < in_size && *prop_pos < prop_size) {
		switch (*prop_pos) {
		case 0:
			if (lzma_lzma_decode_properties(options, in[*in_pos]))
				return LZMA_HEADER_ERROR;
			break;

		case 1:
			// Check that reserved bits are unset.
			if (in[*in_pos] & 0xC0)
				return LZMA_HEADER_ERROR;

			// Set the dictionary size.
			if (in[*in_pos] == 0) {
				options->dictionary_size = 1;

			} else if (in[*in_pos] > 59) {
				// Dictionary size is over 1 GiB. They
				// aren't supported at the moment.
				return LZMA_HEADER_ERROR;
#			if LZMA_DICTIONARY_SIZE_MAX != UINT32_C(1) << 30
#				error Update the if-condition a few lines
#				error above to match LZMA_DICTIONARY_SIZE_MAX.
#			endif

			} else {
				options->dictionary_size
						= (2 | ((in[*in_pos] + 1) & 1))
						<< ((in[*in_pos] - 1)  / 2);
			}
		}

		++*in_pos;
		++*prop_pos;
	}

	return *prop_pos == 2 ? LZMA_STREAM_END : LZMA_OK;
}


static lzma_ret
properties_simple(lzma_options_simple *options, size_t *prop_pos,
		const uint8_t *in, size_t *in_pos, size_t in_size)
{
	if (*prop_pos == 0)
		options->start_offset = 0;

	while (*prop_pos < 4 && *in_pos < in_size)
		options->start_offset |= (uint32_t)(in[(*in_pos)++])
				<< (8 * (*prop_pos)++);

	return *prop_pos == 4 ? LZMA_STREAM_END : LZMA_OK;
}


static lzma_ret
filter_flags_decode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out lzma_attribute((unused)),
		size_t *restrict out_pos lzma_attribute((unused)),
		size_t out_size lzma_attribute((unused)),
		lzma_action action lzma_attribute((unused)))
{
	while (*in_pos < in_size)
	switch (coder->sequence) {
	case SEQ_MISC:
		// Determine the Filter ID and Size of Filter Properties.
		if (in[*in_pos] >= 0xE0) {
			// Using External ID. Prepare the ID
			// for variable-length integer parsing.
			coder->options->id = 0;

			if (in[*in_pos] == 0xFF) {
				// Mark that Size of Filter Properties is
				// unknown, so we know later that there is
				// external Size of Filter Properties present.
				coder->properties_size = UINT64_MAX;
			} else {
				// Take Size of Filter Properties from Misc.
				coder->properties_size = in[*in_pos] - 0xE0;
			}

			coder->sequence = SEQ_ID;

		} else {
			// The Filter ID is the same as Misc.
			coder->options->id = in[*in_pos];

			// The Size of Filter Properties can be calculated
			// from Misc too.
			coder->properties_size = in[*in_pos] / 0x20;

			coder->sequence = SEQ_PROPERTIES;
		}

		++*in_pos;
		break;

	case SEQ_ID: {
		const lzma_ret ret = lzma_vli_decode(&coder->options->id,
				&coder->pos, in, in_pos, in_size);
		if (ret != LZMA_STREAM_END)
			return ret;

		if (coder->properties_size == UINT64_MAX) {
			// We have also external Size of Filter
			// Properties. Prepare the size for
			// variable-length integer parsing.
			coder->properties_size = 0;
			coder->sequence = SEQ_SIZE;
		} else {
			coder->sequence = SEQ_PROPERTIES;
		}

		// Reset pos for its next job.
		coder->pos = 0;
		break;
	}

	case SEQ_SIZE: {
		const lzma_ret ret = lzma_vli_decode(&coder->properties_size,
				&coder->pos, in, in_pos, in_size);
		if (ret != LZMA_STREAM_END)
			return ret;

		coder->pos = 0;
		coder->sequence = SEQ_PROPERTIES;
		break;
	}

	case SEQ_PROPERTIES: {
		lzma_ret ret;

		// Call the filter-specific Properties decoder.
		switch (coder->options->id) {
		case LZMA_FILTER_SPARSE:
			if (coder->properties_size == 0) {
				// No properties, Sparse will repeat nul bytes.
				ret = LZMA_STREAM_END;
				break;
			}

			// Check that the Size of Filter Properties is sane.
			// FIXME
			if (coder->properties_size
					> LZMA_FILTER_PROPERTIES_SIZE_MAX)
				return LZMA_HEADER_ERROR;;

			// Allocate memory for filter options i.e. the bytes
			// that Sparse filter ought to repeat.
			if (coder->options->options == NULL) {
				coder->options->options = lzma_alloc(
						coder->properties_size,
						allocator);
				if (coder->options->options == NULL)
					return LZMA_MEM_ERROR;
			}

			// Copy the properties.
			bufcpy(in, in_pos, in_size, coder->options->options,
					&coder->pos, coder->properties_size);

			// Check if all the properties got copied.
			ret = coder->properties_size == coder->pos
					? LZMA_STREAM_END
					: LZMA_OK;
			break;

		case LZMA_FILTER_LZMA:
			// Allocate memory for filter options if not
			// already allocated.
			if (coder->options->options == NULL) {
				coder->options->options = lzma_alloc(
						sizeof(lzma_options_lzma),
						allocator);
				if (coder->options->options == NULL)
					return LZMA_MEM_ERROR;
			}
			ret = properties_lzma(coder->options->options,
					&coder->pos, coder->properties_size,
					in, in_pos, in_size);
			break;

		case LZMA_FILTER_SUBBLOCK:
			if (coder->properties_size != 0)
				return LZMA_HEADER_ERROR;

			coder->options->options = lzma_alloc(
					sizeof(lzma_options_subblock),
					allocator);
			if (coder->options->options == NULL)
				return LZMA_MEM_ERROR;

			((lzma_options_subblock *)(coder->options->options))
					->allow_subfilters = true;
			ret = LZMA_STREAM_END;
			break;

		case LZMA_FILTER_COPY:
			ret = coder->properties_size > 0
					? LZMA_HEADER_ERROR
					: LZMA_STREAM_END;
			break;

		case LZMA_FILTER_X86:
		case LZMA_FILTER_POWERPC:
		case LZMA_FILTER_IA64:
		case LZMA_FILTER_ARM:
		case LZMA_FILTER_ARMTHUMB:
		case LZMA_FILTER_SPARC:
			if (coder->properties_size == 0) {
				ret = LZMA_STREAM_END;
			} else {
				coder->options->options = lzma_alloc(
						sizeof(lzma_options_simple),
						allocator);
				if (coder->options->options == NULL)
					return LZMA_MEM_ERROR;

				ret = properties_simple(
						coder->options->options,
						&coder->pos,
						in, in_pos, in_size);
			}
			break;

		default:
			return LZMA_HEADER_ERROR;
		}

		return ret;
	}

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static void
filter_flags_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_filter_flags_decoder_init(lzma_next_coder *next,
		lzma_allocator *allocator, lzma_options_filter *options)
{
	if (next->init != &lzma_filter_flags_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;
	}

	options->id = 0;
	options->options = NULL;

	next->coder->options = options;
	next->coder->sequence = SEQ_MISC;
	next->coder->pos = 0;
	next->coder->properties_size = 0;

	next->init = &lzma_filter_flags_decoder_init;
	next->code = &filter_flags_decode;
	next->end = &filter_flags_decoder_end;

	return LZMA_OK;
}


extern LZMA_API lzma_ret
lzma_filter_flags_decoder(lzma_stream *strm, lzma_options_filter *options)
{
	lzma_ret ret;

	ret = lzma_strm_init(strm);
	if (ret != LZMA_OK)
		return ret;

	strm->internal->supported_actions[LZMA_RUN] = true;

	ret = lzma_filter_flags_decoder_init(
			&strm->internal->next, strm->allocator, options);
	if (ret != LZMA_OK)
		lzma_end(strm);

	return ret;
}
