/**
 * \file        lzma.h
 * \brief       The public API of liblzma
 *
 * liblzma is a LZMA compression library with a zlib-like API.
 * liblzma is based on LZMA SDK found from http://7-zip.org/sdk.html.
 *
 * \author      Copyright (C) 1999-2006 Igor Pavlov
 * \author      Copyright (C) 2007 Lasse Collin
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#ifndef LZMA_H
#define LZMA_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * GCC extensions are used conditionally in the public API. It doesn't
 * break anything if these are sometimes enabled and sometimes not, only
 * affects warnings and optimizations.
 */
#if defined(__GNUC__) && __GNUC__ >= 3
#	ifndef lzma_attribute
#		define lzma_attribute(attr) __attribute__(attr)
#	endif
#	ifndef lzma_restrict
#		define lzma_restrict __restrict__
#	endif
#else
#	ifndef lzma_attribute
#		define lzma_attribute(attr)
#	endif
#	ifndef lzma_restrict
#		define lzma_restrict
#	endif
#endif


/***********
 * Headers *
 ***********/

/* size_t */
#include <sys/types.h>

/* NULL */
#include <stddef.h>

/* uint8_t, uint32_t, uint64_t, UINT32_C, UINT64_C, UINT64_MAX. */
#include <inttypes.h>


/******************
 * Initialization *
 ******************/

/**
 * \brief       Initialize all internal static variables
 *
 * Depending on the build options, liblzma may have some internal static
 * variables, that must be initialized before using any other part of
 * the library (*). It is recommended to do these initializations in the very
 * beginning of the application by calling appropriate initialization function.
 *
 * (*) There are some exceptions to this rule. FIXME
 *
 * The initialization functions are not necessarily thread-safe, thus the
 * required initiazations must be done before creating any threads. (The
 * rest of the functions of liblzma are thread-safe.) Calling the
 * initialization functions multiple times does no harm, although it
 * still shouldn't be done when there are multiple threads running.
 *
 * lzma_init_all() initializes all internal static variables by calling
 * lzma_lzma_init_encoder() and lzma_init_decoder().
 *
 * If you need only encoder, decoder, or neither-encoder-nor-decoder
 * functions, you may use other initialization functions, which initialize
 * only a subset of liblzma's internal static variables. Using those
 * functions have the following advantages:
 *  - When linking statically against liblzma, less useless functions will
 *    get linked into the binary. E.g. if you need only the decoder functions,
 *    using lzma_init_decoder() avoids linking bunch of encoder related code.
 *  - There is less things to initialize, making the initialization
 *    process slightly faster.
 */
extern void lzma_init(void);


/**
 * \brief       Initialize internal static variables needed by encoders
 *
 * If you need only the encoder functions, you may use this function to
 * initialize only the things required by encoders.
 *
 * This function also calls lzma_init_check().
 */
extern void lzma_init_encoder(void);


/**
 * \brief       Initialize internal static variables needed by decoders
 *
 * If you need only the decoder functions, you may use this function to
 * initialize only the things required by decoders.
 *
 * This function also calls lzma_init_check().
 */
extern void lzma_init_decoder(void);


/**
 * \brief       Initialize internal static variables needed by integrity checks
 *
 * Currently this initializes CRC32 and CRC64 lookup tables if precalculated
 * tables haven't been built into the library. This function can be useful
 * if the only thing you need from liblzma is the integrity check functions.
 */
extern void lzma_init_check(void);


/**************************
 * Basic type definitions *
 **************************/

/**
 * \brief       Boolean
 *
 * This is here because C89 doesn't have stdbool.h. To set a value for
 * variables having type lzma_bool, you can use
 *   - C99's `true' and `false' from stdbool.h;
 *   - C++'s internal `true' and `false'; or
 *   - integers one (true) and zero (false).
 */
typedef unsigned char lzma_bool;


/**
 * \brief       Return values used by several functions in liblzma
 *
 * Check the descriptions of specific functions to find out which return
 * values they can return and the exact meanings of the values in every
 * situation. The descriptions given here are only suggestive.
 */
typedef enum {
	LZMA_OK                 =  0,
		/**<
		 * \brief       Operation completed successfully
		 */

	LZMA_STREAM_END         =  1,
		/**<
		 * \brief       End of stream was reached
		 *
		 * The application should pick the last remaining output
		 * bytes from strm->next_out. Appropriate _init() or _end()
		 * should be called next.
		 */

	LZMA_STREAM_ERROR       = -2,
		/**<
		 * \brief       Generic error code
		 *
		 * This indicates that the arguments given to the function are
		 * invalid or the internal state of the decoder is corrupt.
		 *   - Function arguments are invalid or the structures
		 *     pointed by the argument pointers are invalid
		 *     e.g. if strm->next_out has been set to NULL when
		 *     calling lzma_code().
		 *   - lzma_* functions have been called in wrong order
		 *     e.g. lzma_code() was called right after lzma_end().
		 *   - If errors occur randomly, the reason might be flaky
		 *     hardware.
		 *
		 * If you think that your code is correct, there can be
		 * a bug in liblzma. See the documentation how to report bugs.
		 */

	LZMA_DATA_ERROR         = -3,
		/**<
		 * \brief       Data is corrupt
		 *
		 * - Encoder: The input size doesn't match the uncompressed
		 *   size given to lzma_*_encoder_init().
		 * - Decoder: The input is corrupt. This includes corrupted
		 *   header, corrupted compressed data, and unmatching
		 *   integrity Check.
		 *
		 * \todo        What can be done if encoder returns this?
		 *              Probably can continue by fixing the input
		 *              amount, but make sure.
		 */

	LZMA_MEM_ERROR          = -4,
		/**<
		 * \brief       Cannot allocate memory
		 *
		 * Memory allocation failed.
		 *
		 * \todo        Make every function to clean up remaining
		 *              allocations automatically if this error occurs.
		 *
		 * \todo        Make sure that it is possible to retry an
		 *              operation if LZMA_MEM_ERROR has occurred.
		 */

	LZMA_BUF_ERROR          = -5,
		/**<
		 * \brief       No progress is possible
		 *
		 * This may happen when avail_in or avail_out is zero.
		 *
		 * \note        This error is not fatal. Coding can continue
		 *              normally once the reason for this error has
		 *              been fixed.
		 */

	LZMA_HEADER_ERROR       = -6,
		/**<
		 * \brief       Invalid or unsupported header
		 *
		 * Invalid or unsupported options, for example
		 *  - unsupported filter(s) or filter options; or
		 *  - reserved bits set in headers (decoder only).
		 *
		 * Rebuilding liblzma with more features enabled, or
		 * upgrading to a newer version of liblzma may help.
		 */

	LZMA_UNSUPPORTED_CHECK  = -7,
		/**<
		 * \brief       Check type is unknown
		 *
		 * The type of Check is not supported, and thus the Check
		 * cannot be calculated. In the encoder, this is an error.
		 * In the decoder, this is only a warning and decoding can
		 * still proceed normally (but the Check is ignored).
		 */

	LZMA_LIMIT_REACHED      = -8
		/**<
		 * \brief       Memory usage limits were reached
		 *
		 * Encoding or decoding was stopped because lzma_index_size_max
		 * or lzma_extra_size_max was reached. You may increase the
		 * limits and try to continue coding. However, in most cases
		 * this is sign of unexpected situation, and indicating an
		 * error to the user could be the most appropriate thing to do.
		 */

} lzma_ret;


/**
 * \brief       The `action' argument for lzma_code()
 */
typedef enum {
	LZMA_RUN      = 0,
		/**<
		 * Processes as much input as possible and as much output
		 * as possible. In the decoder, this action provides best
		 * throughput, but usually will introduce latency.
		 */

	LZMA_FLUSH    = 1,
		/**<
		 * - Multi-Block Stream encoder: All the input data for the
		 *   current Block has been given to the libary (the last
		 *   bytes can still be pending in next_in). Call lzma_code()
		 *   with LZMA_FLUSH until it returns LZMA_STREAM_END. Then
		 *   continue normally with LZMA_RUN or close the Stream
		 *   with LZMA_FINISH.
		 * - Other encoders: Not supported, don't use. Will return
		 *   LZMA_STREAM_ERROR.
		 * - Decoder: Decode only as much as is needed to fill
		 *   next_out. Use this flag when it is likely that you don't
		 *   need more output soon than there is space in next_out.
		 */

	LZMA_FINISH   = 2
		/**<
		 * - Encoder: All the input data has been given to the
		 *   encoder (the last bytes can still be pending in next_in).
		 *   Call lzma_code() with LZMA_FINISH until it returns
		 *   LZMA_STREAM_END.
		 * - Decoder: Not supported, don't use. Will return
		 *   LZMA_STREAM_ERROR.
		 */
} lzma_action;


/*********************
 * Memory management *
 *********************/

/**
 * \brief       Opaque data type used with memory usage limitting functions
 */
typedef struct lzma_memory_limitter_s lzma_memory_limitter;


/**
 * \brief       Allocates and initializes a new lzma_memory_limitter structure
 *
 * It is easy to make liblzma to use huge amounts of memory. This can
 * be a problem especially with the decoder, since it a file requiring
 * huge amounts of memory to uncompress could allow even a denial of
 * service attack if the memory usage wasn't limited.
 *
 * liblzma provides a set of functions to control memory usage. Pointers
 * to these functions can be used in lzma_allocator structure, which makes
 * it easy to limit memory usage with liblzma.
 *
 * The memory limitter functions are not tied to limitting memory usage
 * with liblzma itself. You can use them with anything you like.
 *
 * In multi-threaded applications, only one thread at once may use the same
 * lzma_memory_limitter structure. If there is a need, this limitation may
 * be removed in future versions without breaking the libary API/ABI.
 *
 * \param       limit   Initial memory usage limit in bytes
 *
 * \return      Pointer to allocated and initialized lzma_memory_limitter
 *              structure. On error, NULL is returned. The reason behind
 *              an error is either that malloc() failed or that the given
 *              limit was so small that it didn't allow allocating even
 *              the lzma_memory_limitter structure itself.
 *
 * \note        Excluding lzma_memory_usage(), the functions whose name begin
 *              lzma_memory_ can be used even if lzma_init() hasn't been
 *              called.
 */
extern lzma_memory_limitter *lzma_memory_limitter_create(size_t limit);


/**
 * \brief       Sets a new memory usage limit
 *
 * \param       mem     Pointer to a lzma_memory_limitter structure returned
 *                      earlier by lzma_memry_limit_create().
 * \param       limit   New memory usage limit
 *
 * The new usage limit may be smaller than the amount of memory currently
 * allocated via *mem: New allocations will fail until enough memory has
 * been freed or a new limit is set, but the existing allocatations will
 * stay untouched.
 */
extern void lzma_memory_set_limit(lzma_memory_limitter *mem, size_t limit);


/**
 * \brief       Gets the current memory usage limit
 */
extern size_t lzma_memory_get_limit(lzma_memory_limitter *mem);


/**
 * \brief       Gets the amount of currently allocated memory
 *
 * \note        This value includes the sizes of some helper structures,
 *              thus it will always be larger than the total number of
 *              bytes allocated via lzma_memory_alloc().
 */
extern size_t lzma_memory_get_used(lzma_memory_limitter *mem);


/**
 * \brief       Allocates memory with malloc() if memory limit allows
 *
 * \param       mem     Pointer to a lzma_memory_limitter structure returned
 *                      earlier by lzma_memry_limit_create().
 * \param       nmemb   Number of elements to allocate. While liblzma always
 *                      sets this to one, this function still takes the
 *                      value of nmemb into account to keep the function
 *                      usable with zlib and libbzip2.
 * \param       size    Size of an element. The amount of memory being
 *                      allocated is nmemb * size bytes.
 *
 * \return      Pointer to memory allocated with malloc(nmemb * size).
 *              On error, NULL is returned.
 *
 * \note        This function assumes that nmemb * size is at maximum of
 *              SIZE_MAX. If it isn't, an overflow will occur resulting
 *              invalid amount of memory being allocated.
 */
extern void *lzma_memory_alloc(
		lzma_memory_limitter *mem, size_t nmemb, size_t size);


/**
 * \brief       Removes the pointer from memory limitting list
 *
 * \param       mem     Pointer to a lzma_memory_limitter structure returned
 *                      earlier by lzma_memry_limit_create().
 * \param       ptr     Pointer returned earlier by lzma_memory_alloc().
 *
 * This function removes ptr from the internal list and decreases the
 * counter of used memory accordingly. The ptr itself isn't freed. This is
 * useful when Extra Records allocated by liblzma using lzma_memory_limitter
 * are needed by the application and must not be freed when the
 * lzma_memory_limitter structure is destroyed.
 */
extern void lzma_memory_detach(lzma_memory_limitter *mem, void *ptr);


/**
 * \brief       Frees memory and updates the memory limit list
 *
 * This is like lzma_memory_detach() but also frees the given pointer.
 */
extern void lzma_memory_free(lzma_memory_limitter *mem, void *ptr);


/**
 * \brief       Frees the memory allocated for and by the memory usage limitter
 *
 * \param       mem             Pointer to memory limitter
 * \param       free_allocated  If this is non-zero, all the memory allocated
 *                              by lzma_memory_alloc() using *mem is also
 *                              freed if it hasn't already been freed with
 *                              lzma_memory_free(). Usually this should be
 *                              set to true.
 */
extern void lzma_memory_limitter_end(
		lzma_memory_limitter *mem, lzma_bool free_allocated);


/**
 * \brief       Custom functions for memory handling
 *
 * A pointer to lzma_allocator may be passed via lzma_stream structure
 * to liblzma. The library will use these functions for memory handling
 * instead of the default malloc() and free().
 *
 * liblzma doesn't make an internal copy of lzma_allocator. Thus, it is
 * OK to change these function pointers in the middle of the coding
 * process, but obviously it must be done carefully to make sure that the
 * replacement `free' can deallocate memory allocated by the earlier
 * `alloc' function(s).
 */
typedef struct {
	/**
	 * \brief       Pointer to custom memory allocation function
	 *
	 * Set this to point to your custom memory allocation function.
	 * It can be useful for example if you want to limit how much
	 * memory liblzma is allowed to use: for this, you may use
	 * a pointer to lzma_memory_alloc().
	 *
	 * If you don't want a custom allocator, but still want
	 * custom free(), set this to NULL and liblzma will use
	 * the standard malloc().
	 *
	 * \param       opaque  lzma_allocator.opaque (see below)
	 * \param       nmemb   Number of elements like in calloc().
	 *                      liblzma will always set nmemb to 1.
	 *                      This argument exists only for
	 *                      compatibility with zlib and libbzip2.
	 * \param       size    Size of an element in bytes.
	 *                      liblzma never sets this to zero.
	 *
	 * \return      Pointer to the beginning of a memory
	 *              block of size <b>nmemb * size</b>, or
	 *              NULL if allocation fails for some reasons.
	 *              When allocation fails, functions of liblzma
	 *              return LZMA_MEM_ERROR.
	 */
	void *(*alloc)(void *opaque, size_t nmemb, size_t size);

	/**
	 * \brief       Pointer to custom memory freeing function
	 *
	 * Set this to point to your custom memory freeing function.
	 * If lzma_memory_alloc() is used as allocator, this should
	 * be set to lzma_memory_free().
	 *
	 * If you don't want a custom freeing function, but still
	 * want a custom allocator, set this to NULL and liblzma
	 * will use the standard free().
	 *
	 * \param       opaque  lzma_allocator.opaque (see below)
	 * \param       ptr     Pointer returned by
	 *                      lzma_allocator.alloc(), or when it
	 *                      is set to NULL, a pointer returned
	 *                      by the standard malloc().
	 */
	void (*free)(void *opaque, void *ptr);

	/**
	 * \brief       Pointer passed to .alloc() and .free()
	 *
	 * opaque is passed as the first argument to lzma_allocator.alloc()
	 * and lzma_allocator.free(). This intended to ease implementing
	 * custom memory allocation functions for use with liblzma.
	 *
	 * When using lzma_memory_alloc() and lzma_memory_free(), opaque
	 * must point to lzma_memory_limitter structure allocated and
	 * initialized with lzma_memory_limitter_create().
	 *
	 * If you don't need this, you should set it to NULL.
	 */
	void *opaque;

} lzma_allocator;


/* NOTE: See also lzma_memory_usage() later in this file. */


/************************************
 * Passing data to and from liblzma *
 ************************************/

/**
 * \brief       Internal data structure
 *
 * The contents of this structure is not visible outside the library.
 */
typedef struct lzma_internal_s lzma_internal;


/**
 * \brief       Passing data to and from liblzma
 *
 * The lzma_stream structure is used for
 *   - passing pointers to input and output buffers to liblzma;
 *   - defining custom memory hander functions; and
 *   - holding a pointer to coder-specific internal data structures.
 *
 * Before calling any of the lzma_*_init() functions the first time,
 * the application must reset lzma_stream to LZMA_STREAM_INIT. The
 * lzma_*_init() function will verify the options, allocate internal
 * data structures and store pointer to them into `internal'. Finally
 * total_in and total_out are reset to zero. In contrast to zlib,
 * next_in and avail_in are ignored by the initialization functions.
 *
 * The actual coding is done with the lzma_code() function. Application
 * must update next_in, avail_in, next_out, and avail_out between
 * calls to lzma_decode() just like with zlib.
 *
 * In contrast to zlib, even the decoder requires that there always
 * is at least one byte space in next_out; if avail_out == 0,
 * LZMA_BUF_ERROR is returned immediatelly. This shouldn't be a problem
 * for most applications that already use zlib, but it's still worth
 * checking your application.
 *
 * Application may modify values of total_in and total_out as it wants.
 * They are updated by liblzma to match the amount of data read and
 * written, but liblzma doesn't use the values internally.
 *
 * Application must not touch the `internal' pointer.
 */
typedef struct {
	uint8_t *next_in;   /**< Pointer to the next input byte. */
	size_t avail_in;    /**< Number of available input bytes in next_in. */
	uint64_t total_in;  /**< Total number of bytes read by liblzma. */

	uint8_t *next_out;  /**< Pointer to the next output position. */
	size_t avail_out;   /**< Amount of free space in next_out. */
	uint64_t total_out; /**< Total number of bytes written by liblzma. */

	/**
	 * Custom memory allocation functions. Set to NULL to use
	 * the standard malloc() and free().
	 */
	lzma_allocator *allocator;

	/** Internal state is not visible to outsiders. */
	lzma_internal *internal;

} lzma_stream;


/**
 * \brief       Initialization for lzma_stream
 *
 * When you declare an instance of lzma_stream, you can immediatelly
 * initialize it so that initialization functions know that no memory
 * has been allocated yet:
 *
 *     lzma_stream strm = LZMA_STREAM_INIT;
 */
#define LZMA_STREAM_INIT { NULL, 0, 0, NULL, 0, 0, NULL, NULL }


/**
 * \brief       Initialization for lzma_stream
 *
 * This is like LZMA_STREAM_INIT, but this can be used when the lzma_stream
 * has already been allocated:
 *
 *     lzma_stream *strm = malloc(sizeof(lzma_stream));
 *     if (strm == NULL)
 *         return LZMA_MEM_ERROR;
 *     *strm = LZMA_STREAM_INIT_VAR;
 */
extern const lzma_stream LZMA_STREAM_INIT_VAR;


/***************************
 * Basic Metadata handling *
 ***************************/

/* Extra Record IDs */

#define LZMA_EXTRA_PADDING      0x00
#define LZMA_EXTRA_OPENPGP      0x01
#define LZMA_EXTRA_FILTERS      0x02
#define LZMA_EXTRA_COMMENT      0x03
#define LZMA_EXTRA_CHECKS       0x04
#define LZMA_EXTRA_FILENAME     0x05
#define LZMA_EXTRA_MTIME        0x07
#define LZMA_EXTRA_MTIME_HR     0x09
#define LZMA_EXTRA_MIME_TYPE    0x0B
#define LZMA_EXTRA_HOMEPAGE     0x0D


/**
 * \brief       Extra Records
 *
 * The .lzma format provides a way to store custom information along
 * the actual compressed content. Information about these Records
 * are passed to and from liblzma via this linked list.
 */
typedef struct lzma_extra_s lzma_extra;
struct lzma_extra_s {
	/**
	 * \brief       Pointer to the next Extra Record
	 *
	 * This is NULL on the last Extra Record.
	 */
	lzma_extra *next;

	/**
	 * \brief       Record ID
	 *
	 * Extra Record IDs are divided in three categories:
	 *   - Zero is a special case used for padding. It doesn't have
	 *     Size of Data fields.
	 *   - Odd IDs (1, 3, 5, ...) are Safe-to-Copy IDs.
	 *     These can be preserved as is if the Stream is
	 *     modified in a way that doesn't alter the actual
	 *     uncompressed content.
	 *   - Even IDs (2, 4, 6, ...) are Unsafe-to-Copy IDs.
	 *     If the .lzma Stream is modified in any way,
	 *     the Extra Records having a sensitive ID should
	 *     be removed or updated accordingly.
	 *
	 * Refer to the .lzma file format specification for
	 * the up to date list of Extra Record IDs.
	 */
	uint64_t id;

	/**
	 * \brief       Size of the Record data
	 *
	 * In case of strings, this should not include the
	 * trailing '\0'.
	 */
	size_t size;

	/**
	 * \brief       Record data
	 *
	 * Record data is often a string in UTF-8 encoding,
	 * but it can be arbitrary binary data. In case of
	 * strings, the trailing '\0' is usually not stored
	 * in the .lzma file.
	 *
	 * To ease working with Extra Records containing strings,
	 * liblzma always adds '\0' to the end of data even when
	 * it wasn't present in the .lzma file. This '\0' is not
	 * counted in the size of the data.
	 */
	uint8_t *data;
};


/**
 * \brief      Maximum number of Index Records in Metadata Block
 *
 * Each Index Record takes sizeof(lzma_index) bytes of memory (usually
 * 20 or 24 bytes). Default limit is 131072 Records, which requires
 * roughly 3 MiB of RAM.
 */
extern size_t lzma_index_size_max;


/**
 * \brief       Maximum memory usage of Extra Records
 *
 * Each Extra Record takes size(lzma_extra) bytes of memory (usually
 * 20-32 bytes) plus size of the data stored to the Extra Record.
 * This setting limits the total amount of memory that Extra Records
 * may consume RAM when decoding Metadata. The default is 8 MiB.
 */
extern size_t lzma_extra_size_max;


/***************************************
 * LZMA-specific options and constants *
 ***************************************/

/**
 * \brief       LZMA compression modes
 *
 * Currently there are only two modes. Earlier LZMA SDKs had also third
 * mode between fast and best.
 */
typedef enum {
	LZMA_MODE_FAST = 0,
		/**<
		 * \brief       Fast compression
		 *
		 * Fast mode is usually at its best when combined with
		 * a hash chain match finder.
		 */

	LZMA_MODE_BEST = 2
		/**<
		 * \brief       Best compression ratio
		 *
		 * This is usually notably slower than fast mode. Use this
		 * together with binary tree match finders to expose the
		 * full potential of the LZMA encoder.
		 */
} lzma_mode;


/**
 * \brief       Match finders
 *
 * Match finder has major effect on both speed and compression ratio.
 * Usually hash chains are faster than binary trees.
 */
typedef enum {
	LZMA_MF_INVALID = -1,
		/**<
		 * \brief       Invalid match finder ID
		 *
		 * Used only as a return value of
		 * lzma_match_finder_name_to_id().
		 */

	LZMA_MF_HC3     = 0x03,
		/**<
		 * \brief       Hash Chain with 3 bytes hashing
		 *
		 * \todo Memory requirements
		 */

	LZMA_MF_HC4     = 0x04,
		/**<
		 * \brief       Hash Chain with 4 bytes hashing
		 *
		 * Memory requirements: 7.5 * dictionary_size + 4 MiB
		 */

	LZMA_MF_BT2     = 0x12,
		/**<
		 * \brief       Binary Tree with 2 bytes hashing
		 *
		 * Memory requirements: 9.5 * dictionary_size + 4 MiB
		 */

	LZMA_MF_BT3     = 0x13,
		/**<
		 * \brief       Binary Tree with 3 bytes hashing
		 *
		 * Memory requirements: 11.5 * dictionary_size + 4 MiB
		 */

	LZMA_MF_BT4     = 0x14
		/**<
		 * \brief       Binary Tree with 4 bytes hashing
		 *
		 * Memory requirements: 11.5 * dictionary_size + 4 MiB
		 */
} lzma_match_finder;


/**************************
 * Filter IDs and options *
 **************************/

/* These are as defines instead of enums, because Filter ID is 64-bit. */

#define LZMA_FILTER_COPY        UINT64_C(0x00)
	/**<
	 * No filtering; the data is copied as is. You usually don't
	 * need to specify this explicitly.
	 */

#define LZMA_FILTER_SUBBLOCK    UINT64_C(0x01)
	/**<
	 * Subblock filter
	 */

#define LZMA_FILTER_SPARSE      UINT64_C(0x02)
	/**<
	 * Sparse filter
	 */

#define LZMA_FILTER_X86         UINT64_C(0x04)
	/**<
	 * BCJ (Branch, Call, Jump) filter for x86 binaries
	 */

#define LZMA_FILTER_POWERPC     UINT64_C(0x05)
	/**<
	 * Filter for Big endian PowerPC binaries
	 */

#define LZMA_FILTER_IA64        UINT64_C(0x06)
	/**<
	 * Filter for IA64 (Itanium) binaries.
	 */

#define LZMA_FILTER_ARM         UINT64_C(0x07)
	/**<
	 * Filter for ARM binaries.
	 */

#define LZMA_FILTER_ARMTHUMB    UINT64_C(0x08)
	/**<
	 * Filter for ARMThumb binaries.
	 */

#define LZMA_FILTER_SPARC       UINT64_C(0x09)
	/**<
	 * Filter for SPARC binaries.
	 */

#define LZMA_FILTER_LZMA        UINT64_C(0x40)
	/**<
	 * LZMA filter (Lempel-Ziv-Markov chain-Algorithm)
	 */


/**
 * \brief       Filter options
 *
 * This structure is used to pass Filter ID and a pointer filter's options
 * to liblzma.
 */
typedef struct {
	/**
	 * \brief       Filter ID
	 *
	 * Use constants whose name begin with `LZMA_FILTER_' to specify
	 * different filters. In an array of lzma_option_filter structures,
	 * use UINT64_MAX to indicate end of filters.
	 */
	uint64_t id;

	/**
	 * \brief       Pointer to filter-specific options structure
	 *
	 * If the filter doesn't need options, set this to NULL. If id is
	 * set to UINT64_MAX, options is ignored.
	 *
	 * Some filters support changing the options in the middle of
	 * the encoding process. These filters store the pointer of the
	 * options structure and communicate with the application via
	 * modifications of the options structure.
	 */
	void *options;

} lzma_options_filter;


/**
 * \brief       Options specific to the LZMA method handler
 *
 * The first five fields are stored to stream header in both
 * LZMA_HEADER_NATIVE and LZMA_HEADER_ALONE header formats.
 * When decoding data encoded using LZMA_HEADER_RAW, the application
 * must supply dictionary_size, literal_context_bits, literal_pos_bits
 * and pos_bits to the decoder via the second argument of
 * lzma_decoder_create() or lzma_decoder_reset().
 */
typedef struct {
	/**********************************
	 * LZMA encoding/decoding options *
	 **********************************/

	/* These options are required in encoder and also with raw decoding. */

	/**
	 * \brief       Dictionary size in bytes
	 *
	 * Dictionary size indicates how many bytes of the recently processed
	 * uncompressed data is kept in memory. One method to reduce size of
	 * the uncompressed data is to store distance-length pairs, which
	 * indicate what data to repeat from the dictionary buffer. Thus,
	 * the bigger the dictionary, the better compression ratio usually is.
	 *
	 * Raw decoding: Too big dictionary does no other harm than
	 * wasting memory. This value is ignored by lzma_raw_decode_buffer(),
	 * because it uses the target buffer as the dictionary.
	 */
	uint32_t dictionary_size;
#	define LZMA_DICTIONARY_SIZE_MIN            1
#	define LZMA_DICTIONARY_SIZE_MAX            (UINT32_C(1) << 30)
#	define LZMA_DICTIONARY_SIZE_DEFAULT        (UINT32_C(1) << 23)

	/**
	 * \brief       Number of literal context bits
	 *
	 * How many of the highest bits of the previous uncompressed
	 * eight-bit byte (also known as `literal') are taken into
	 * account when predicting the bits of the next literal.
	 *
	 * \todo        Example
	 */
	uint32_t literal_context_bits;
#	define LZMA_LITERAL_CONTEXT_BITS_MIN       0
#	define LZMA_LITERAL_CONTEXT_BITS_MAX       8
#	define LZMA_LITERAL_CONTEXT_BITS_DEFAULT   3

	/**
	 * \brief       Number of literal position bits
	 *
	 * How many of the lowest bits of the current position (number
	 * of bytes from the beginning of the uncompressed data) in the
	 * uncompressed data is taken into account when predicting the
	 * bits of the next literal (a single eight-bit byte).
	 *
	 * \todo        Example
	 */
	uint32_t literal_pos_bits;
#	define LZMA_LITERAL_POS_BITS_MIN           0
#	define LZMA_LITERAL_POS_BITS_MAX           4
#	define LZMA_LITERAL_POS_BITS_DEFAULT       0

	/**
	 * \brief       Number of position bits
	 *
	 * How many of the lowest bits of the current position in the
	 * uncompressed data is taken into account when estimating
	 * probabilities of matches. A match is a sequence of bytes for
	 * which a matching sequence is found from the dictionary and
	 * thus can be stored as distance-length pair.
	 *
	 * Example: If most of the matches occur at byte positions
	 * of 8 * n + 3, that is, 3, 11, 19, ... set pos_bits to 3,
	 * because 2**3 == 8.
	 */
	uint32_t pos_bits;
#	define LZMA_POS_BITS_MIN                   0
#	define LZMA_POS_BITS_MAX                   4
#	define LZMA_POS_BITS_DEFAULT               2


	/******************************************
	 * LZMA options needed only when encoding *
	 ******************************************/

	/** LZMA compression mode */
	lzma_mode mode;

	/**
	 * \brief       Number of fast bytes
	 *
	 * Number of fast bytes determines how many bytes the encoder
	 * compares from the match candidates when looking for the best
	 * match. Bigger fast bytes value usually increase both compression
	 * ratio and time.
	 */
	uint32_t fast_bytes;
#	define LZMA_FAST_BYTES_MIN                 5
#	define LZMA_FAST_BYTES_MAX                 273
#	define LZMA_FAST_BYTES_DEFAULT             128

	/** Match finder ID */
	lzma_match_finder match_finder;

	/**
	 * \brief       Match finder cycles
	 *
	 * Higher values give slightly better compression ratio but
	 * decrease speed. Use special value 0 to let liblzma use
	 * match-finder-dependent default value.
	 *
	 * \todo        Write much better description.
	 */
	uint32_t match_finder_cycles;

} lzma_options_lzma;


typedef struct {
	/**
	 * \brief       Size of the buffer to be repeated
	 *
	 * If this is zero, the Sparse filter will repeat '\0' bytes.
	 */
	size_t size;
#	define LZMA_SPARSE_SIZE_MAX 65536

	/**
	 * \brief       Bytes to repeat
	 *
	 * If `size' is zero, the encoder doesn't read `bytes'.
	 *
	 * The initialization function of the Sparse filter makes internal
	 * copy of these bytes, so there is no need for the application
	 * to keep the options of the Sparse filter available after
	 * initialization.
	 */
	uint8_t *bytes;

} lzma_options_sparse;


/**
 * \brief       Options for the Subblock filter
 *
 * Specifying options for the Subblock filter is optional: if the pointer
 * options is NULL, no subfilters are allowed and the default value is used
 * for subblock_data_size.
 */
typedef struct {
	/* Options for encoder and decoder */

	/**
	 * \brief       Allowing subfilters
	 *
	 * If this true, subfilters are allowed.
	 *
	 * This option is used in both encoder and decoder.
	 */
	lzma_bool allow_subfilters;

	/* Options for encoder only */

	/**
	 * \brief       Flush flag
	 *
	 * Set this to true to flush the current Subblock Data to output.
	 * After flushing, the variables subblock_data_size and subfilter
	 * are read from this structure. As the last step, liblzma sets
	 * the flush variable back to false. This way the application can
	 * detect when the old Subfilter is no longer in use.
	 */
	lzma_bool flush;

	/**
	 * \brief       Size of the Subblock Data part of each Subblock
	 *
	 * Bigger values
	 *   - save a few bytes of space;
	 *   - increase latency in the encoder (but no effect for decoding);
	 *   - decrease memory locality (increased cache pollution) in the
	 *     encoder (no effect in decoding).
	 */
	uint32_t subblock_data_size;
#	define LZMA_SUBBLOCK_DATA_SIZE_MIN 1
#	define LZMA_SUBBLOCK_DATA_SIZE_MAX ((UINT32_C(1) << 28) - 1)
#	define LZMA_SUBBLOCK_DATA_SIZE_DEFAULT 4096

	/**
	 * \brief       Subfilter and its options
	 *
	 * When no Subfilter is used, the data is copied as is into Subblocks.
	 * Setting a Subfilter allows encoding some parts of the data with
	 * an additional filter. It is possible to many different Subfilters
	 * in the same Block, although only one can be used at once.
	 */
	lzma_options_filter subfilter;

	/**
	 * \brief       Size of the data being fed to the Subfilter
	 *
	 * If the amount of data to be processed with the Subfilter is known
	 * beforehand, it is possible to tell it via this variable. When
	 * Uncompressed Size of the Subfilter is unknown, this must be set
	 * to UINT64_MAX.
	 */
	uint64_t subfilter_uncompressed_size;

} lzma_options_subblock;


/**
 * \brief       Options for so called "simple" filters
 *
 * The simple filters never change the size of the data. Specifying options
 * for them is optional: if pointer to options is NULL, default values are
 * used.
 */
typedef struct {
	/**
	 * \brief       Start offset for branch conversions
	 *
	 * You very probably have no use for this. Simply set the options
	 * pointer to NULL and be happy.
	 *
	 * This setting is useful only when the same filter is used
	 * _separately_ for multiple sections of the same executable file,
	 * and the sections contain cross-section branch/call/jump
	 * instructions. In that case it is benefical to set the start
	 * offset of the non-first sections so that the relative addresses
	 * of the cross-section branch/call/jump instructions will use the
	 * same absolute addresses as in the first section.
	 *
	 * When the pointer to options is NULL, the default value is used.
	 * The default value is zero.
	 */
	uint32_t start_offset;

} lzma_options_simple;


/***************************
 * Common type definitions *
 ***************************/

/**
 * \brief       Type of the Check
 *
 * The .lzma format supports multiple types of Checks that are calculated
 * from the uncompressed data (unless it is empty; then it's calculated
 * from Block Header).
 */
typedef enum {
	LZMA_CHECK_NONE     = 0,
		/**<
		 * No Check is calculated.
		 *
		 * Size of the Check field: 0 bytes
		 */

	LZMA_CHECK_CRC32    = 1,
		/**<
		 * CRC32 using the polynomial from the IEEE 802.3 standard
		 *
		 * Size of the Check field: 4 bytes
		 */

	LZMA_CHECK_CRC64    = 3,
		/**<
		 * CRC64 using the polynomial from the ECMA-182 standard
		 *
		 * Size of the Check field: 8 bytes
		 */

	LZMA_CHECK_SHA256   = 5
		/**<
		 * SHA-256
		 *
		 * Size of the Check field: 32 bytes
		 */
} lzma_check_type;


/**
 * \brief       Maximum valid Check ID
 *
 * The .lzma file format specification specifies eight Check IDs (0-7). Some
 * of them are only reserved i.e. no actual Check algorithm has been assigned.
 * Still liblzma accepts any of these eight IDs for future compatibility
 * when decoding files. If a valid but unsupported Check ID is detected,
 * liblzma indicates a warning with LZMA_UNSUPPORTED_CHECK.
 *
 * FIXME bad desc
 */
#define LZMA_CHECK_ID_MAX 7


/**
 * \brief       Options for .lzma Stream encoder
 */
typedef struct {
	/**
	 * \brief       Type of integrity Check
	 *
	 * The type of the integrity Check is stored into Stream Header
	 * and Stream Footer. The same Check is used for all Blocks in
	 * the Stream.
	 */
	lzma_check_type check;

	/**
	 * \brief       Precense of CRC32 of the Block Header
	 *
	 * Set this to true if CRC32 of every Block Header should be
	 * calculated and stored in the Block Header. This is recommended.
	 *
	 * This setting is stored into Stream Header and Stream Footer.
	 */
	lzma_bool has_crc32;

	/**
	 * \brief       Alignment of the beginning of the Stream
	 *
	 * Certain filters handle data in bigger chunks than single bytes.
	 * This affects two things:
	 *   - Performance: aligned memory access is usually faster.
	 *   - Further compression ratio in custom file formats: if you
	 *     encode multiple Blocks with some non-compression filter
	 *     such as LZMA_FILTER_POWERPC, it is a good idea to keep
	 *     the inter-Block alignment correct to maximize compression
	 *     ratio when all these Blocks are finally compressed as a
	 *     single step.
	 *
	 * Usually the Stream is stored into its own file, thus
	 * the alignment is usually zero.
	 */
	uint32_t alignment;

	/**
	 * \brief       Array of filters used to encode Data Blocks
	 *
	 * There can be at maximum of seven filters. The end of the array
	 * is marked with .id = UINT64_MAX. (That's why the array has eight
	 * members.) Minimum number of filters is zero; in that case, an
	 * implicit Copy filter is used.
	 */
	lzma_options_filter filters[8];

	/**
	 * \brief       Array of filters used to encode Metadata Blocks
	 *
	 * This is like filters[] but for Metadata Blocks. If Metadata
	 * Blocks are compressed, they usually are compressed with
	 * settings that require only little memory to uncompress e.g.
	 * LZMA with 64 KiB dictionary.
	 *
	 * \todo        Recommend a preset.
	 *
	 * When liblzma sees that the Metadata Block would be very small
	 * even in uncompressed form, it is not compressed no matter
	 * what filter have been set here. This is to avoid possibly
	 * increasing the size of the Metadata Block with bad compression,
	 * and to avoid useless overhead of filters in uncompression phase.
	 */
	lzma_options_filter metadata_filters[8];

	/**
	 * \brief       Extra information in the Header Metadata Block
	 */
	lzma_extra *header;

	/**
	 * \brief       Extra information in the Footer Metadata Block
	 *
	 * It is enough to set this pointer any time before calling
	 * lzma_code() with LZMA_FINISH as the second argument.
	 */
	lzma_extra *footer;

} lzma_options_stream;


/**
 * \brief       Options for files in the LZMA_Alone format
 */
typedef struct {
	/**
	 * \brief       Uncompressed Size and usage of End of Payload Marker
	 *
	 * In contrast to .lzma Blocks, LZMA_Alone format cannot have both
	 * uncompressed size field in the header and end of payload marker.
	 * If you don't know the uncompressed size beforehand, set it to
	 * UINT64_MAX and liblzma will embed end of payload marker.
	 */
	uint64_t uncompressed_size;

	/**
	 * \brief       LZMA options
	 *
	 * The LZMA_Alone format supports only one filter: the LZMA filter.
	 *
	 * \note        There exists also an undocumented variant of the
	 *              LZMA_Alone format, which uses the x86 filter in
	 *              addition to LZMA. This format was never supported
	 *              by LZMA Utils and is not supported by liblzma either.
	 */
	lzma_options_lzma lzma;

} lzma_options_alone;


/********************
 * Coding functions *
 ********************/

/**
 * \brief       Encodes or decodes data
 *
 * Once the lzma_stream has been successfully initialized with one of
 * lzma_*_init() functions, the actual encoding or decoding is done
 * using this function.
 *
 * \return      - LZMA_OK: So far all good.
 *              - LZMA_STREAM_END:
 *                  - Encoder: LZMA_FLUSH or LZMA_FINISH completed.
 *                  - Decoder: End of uncompressed data was reached.
 *              - LZMA_BUF_ERROR: Unable to progress. Provide more input or
 *                output space, and call this function again.
 *              - LZMA_MEM_ERROR: Unable to allocate memory. Free some
 *                memory and call this function again.
 *              - LZMA_DATA_ERROR:
 *                  - Encoder: Filter(s) cannot process the given data.
 *                  - Decoder: Compressed data is corrupt.
 *              - LZMA_HEADER_ERROR:
 *                  - Encoder: Invalid options
 *                  - Decoder: Some header is in wrong format or corrupt.
 *              - LZMA_STREAM_ERROR: Invalid arguments or internal state
 *                of the coder is corrupt.
 */
extern lzma_ret lzma_code(lzma_stream *strm, lzma_action action);


/**
 * \brief       Frees memory allocated for the coder data structures
 *
 * \param       strm    Pointer to lzma_stream that is at least initialized
 *                      with LZMA_STREAM_INIT.
 *
 * \return      Possible return values:
 *                - LZMA_OK: Memory freed successfully or everything
 *                  was already freed.
 *                - LZMA_STREAM_ERROR: strm was NULL or internal state
 *                  was inconsistent.
 *
 * \note        zlib indicates an error if application end()s unfinished
 *              stream. liblzma doesn't do this, and assumes that
 *              application knows what it is doing.
 */
extern lzma_ret lzma_end(lzma_stream *strm);


/****************************
 * Initialization functions *
 ****************************/

/* Encoders */

/**
 * \brief       Initializes Single-Block .lzma Stream encoder
 *
 * This is the function that most developers are looking for. :-) It
 * compresses using the specified options without storing any extra
 * information.
 *
 * \todo        Describe that is_metadata is ignored, maybe something else.
 */
extern lzma_ret lzma_stream_encoder_single(
		lzma_stream *strm, const lzma_options_stream *options);


/**
 * \brief       Initializes Multi-Block .lzma Stream encoder
 *
 */
extern lzma_ret lzma_stream_encoder_multi(
		lzma_stream *strm, const lzma_options_stream *options);


/**
 * \brief       Initializes LZMA_Alone encoder
 *
 * LZMA_Alone files have the suffix .lzma like the .lzma Stream files.
 * LZMA_Alone format supports only one filter, the LZMA filter. There is
 * no support for integrity checks like CRC32.
 *
 * Use this format if and only if you need to create files readable by
 * legacy LZMA tools.
 */
extern lzma_ret lzma_alone_encoder(
		lzma_stream *strm, const lzma_options_alone *options);


/* Decoders */

/**
 * \brief       Decode .lzma Streams and LZMA_Alone files with autodetection
 *
 * Autodetects between the .lzma Stream and LZMA_Alone formats, and
 * calls lzma_stream_decoder_init() or lzma_alone_decoder_init() once
 * the type of the file has been detected.
 *
 * \param       strm        Pointer to propertily prepared lzma_stream
 * \param       header      Pointer to hold a pointer to Extra Records read
 *                          from the Header Metadata Block. Use NULL if
 *                          you don't care about Extra Records.
 * \param       footer      Same as header, but for Footer Metadata Block.
 *
 * \return      - LZMA_OK: Initialization was successful.
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 *
 * \note        lzma_extra_size_max limits memory usage when header or footer
 *              is not NULL. The limit applies independently for header and
 *              footer. When header or footer is NULL, the Extra Records
 *              are parsed but not stored, thus lzma_extra_size_max doesn't
 *              apply.
 */
extern lzma_ret lzma_auto_decoder(lzma_stream *strm,
		lzma_extra **header, lzma_extra **footer);


/**
 * \brief       Initializes decoder for .lzma Stream
 *
 * \param       strm        Pointer to propertily prepared lzma_stream
 * \param       header      Pointer to hold a pointer to Extra Records read
 *                          from the Header Metadata Block. Use NULL if
 *                          you don't care about Extra Records.
 * \param       footer      Same as header, but for Footer Metadata Block.
 *
 * \return      - LZMA_OK: Initialization was successful.
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 *
 * \note        See the note in lzma_auto_decoder().
 */
extern lzma_ret lzma_stream_decoder(lzma_stream *strm,
		lzma_extra **header, lzma_extra **footer);


/**
 * \brief       Initializes decoder for LZMA_Alone file
 *
 * \param       strm        Pointer to propertily prepared lzma_stream
 *
 * \return      - LZMA_OK: Initialization was successful.
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 */
extern lzma_ret lzma_alone_decoder(lzma_stream *strm);


/***********
 * Presets *
 ***********/

/**
 * \brief       Table of encoder presets
 *
 * lzma_presets[0] is the fastest and lzma_preset[9] is the slowest.
 * The values are subject to changes between liblzma versions.
 */
/* extern const lzma_options_block lzma_presets[10]; */


/**
 * \brief       Table of presets for the LZMA filter
 *
 * lzma_presets[0] is the fastest and lzma_preset[8] is the slowest.
 * These presets match the switches -1 .. -9 of the lzma command line tool
 *
 * The preset values are subject to changes between liblzma versions.
 */
extern const lzma_options_lzma lzma_preset_lzma[9];


/****************
 * Memory usage *
 ****************/

/**
 * \brief       Calculate rough memory requirements for given filter chain
 *
 * \param       filters     Array of filters terminated with .id == UINT64_MAX.
 * \param       is_encoder  Set to true when calculating memory requirements
 *                          of an encoder; false for decoder.
 *
 * \return      Number of mebibytes (MiB i.e. 2^20) required for the given
 *              encoder or decoder filter chain.
 *
 * \note        If calculating memory requirements of encoder, lzma_init() or
 *              lzma_init_encoder() must have been called earlier. Similarly,
 *              if calculating memory requirements of decoder, lzma_init() or
 *              lzma_init_decoder() must have been called earlier.
 */
extern uint32_t lzma_memory_usage(
		const lzma_options_filter *filters, lzma_bool is_encoder);


/******************
 * Version number *
 ******************/

/**
 * \brief       Returns versions number of liblzma as a string
 *
 * This function may be useful if you want to display which version of
 * libilzma your application is currently using.
 *
 * \return      Returns a pointer to a statically allocated string constant,
 *              which contains the version number of liblzma. The format of
 *              the version string is usually (but not necessarily) x.y.z
 *              e.g. "4.42.1". Alpha and beta versions contain a suffix
 *              ("4.42.0alpha").
 */
extern const char *lzma_version_string(void);


/**
 * \brief       Returns version number of liblzma as an integer
 *
 * This function may be useful if you need to get the version number
 * of liblzma in an easily comparable form.
 *
 * \todo        Decide and describe the format of the returned integer.
 */
extern uint32_t lzma_version_number(void);


#ifdef __cplusplus
}
#endif

#endif /* ifndef LZMA_H */
