/**
 * \file        lzma_adv.h
 * \brief       Advanced features of liblzma
 *
 * These functions are used by liblzma internally by the functions declared
 * in lzma.h. In some situations low-level access to .lzma format is needed,
 * so some of the originally internal functions have been made public. The
 * API and ABI of these functions is as stable as the ABI of the functions
 * declared in lzma.h.
 *
 * \author      Copyright (C) 1999-2006 Igor Pavlov
 * \author      Copyright (C) 2007 Lasse Collin
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#ifndef LZMA_ADV_H
#define LZMA_ADV_H

#include "lzma.h"

#ifdef __cplusplus
extern "C" {
#endif

/******************************
 * Advanced Metadata handling *
 ******************************/

/**
 * \brief       Index Records
 *
 * Every Multi-Block Stream has Index at least in one Metadata Block.
 * If it is in both Metadata Blocks, the Indexes must be identical.
 */
typedef struct lzma_index_s lzma_index;
struct lzma_index_s {
	/**
	 * \brief       Pointer to the next Index Record
	 *
	 * This is NULL on the last Index Record.
	 */
	lzma_index *next;

	/**
	 * \brief       Total Size of the respective Block
	 *
	 * Total Size includes Block Header, Compressed Data, and
	 * Block Footer.
	 */
	uint64_t total_size;

	/**
	 * \brief       Uncompressed Size of the respective Block
	 */
	uint64_t uncompressed_size;
};


/**
 * \brief       Information stored into a Metadata Block
 *
 * This structure holds all the information that can be stored to
 * a Metadata Block.
 */
typedef struct {
	/**
	 * \brief       Size of Header Metadata Block
	 */
	uint64_t header_metadata_size;

	/**
	 * \brief       Total Size of the Stream
	 */
	uint64_t total_size;

	/**
	 * \brief       Uncompressed Size of the Stream
	 */
	uint64_t uncompressed_size;

	/**
	 * \brief       Index of the Blocks stored in the Stream
	 */
	lzma_index *index;

	/**
	 * \brief       Extra information
	 */
	lzma_extra *extra;

} lzma_metadata;


/**
 * Opaque type for Metadata encoder
 */
typedef struct lzma_metadata_encoder_s lzma_metadata_encoder;


/**
 * Opaque type for Metadata decoder
 */
typedef struct lzma_metadata_decoder_s lzma_metadata_decoder;


/**
 * \brief       Initializes Metadata encoder
 *
 * \param       coder       Pointer to a pointer to hold Metadata encoder's
 *                          internal state. Original value is ignored, thus
 *                          you don't need to initialize the pointer.
 * \param       allocator   Custom memory allocator; usually NULL.
 * \param       metadata    Pointer to Metadata to encoded
 *
 * \return      - LZMA_OK: Initialization succeeded.
 *              - LZMA_MEM_ERROR: Cannot allocate memory for *coder.
 *
 * The initialization function makes internal copy of the *metadata structure.
 * However, the linked lists metadata->index and metadata->extra are NOT
 * copied. Thus, the application may destroy *metadata after initialization
 * if it likes, but not Index or Extra.
 */
extern lzma_ret lzma_metadata_encoder_init(lzma_metadata_encoder **coder,
		lzma_allocator *allocator, const lzma_metadata *metadata);


/**
 * \brief       Encodes Metadata
 *
 * \param       coder       Metadata coder initialized with
 *                          lzma_metadata_encoder_init().
 * \param       strm_out    FIXME
 *
 * \note        When lzma_allocator is needed, strm->allocator is used.
 */
extern lzma_ret lzma_metadata_encode(
		lzma_metadata_encoder *coder, lzma_stream *strm_out);


/**
 * \brief       Frees memory allocated for Metadata encoder
 */
extern void lzma_metadata_encoder_end(
		lzma_metadata_encoder **coder, lzma_allocator *allocator);


/**
 * \brief       Calculate the encoded size of Metadata
 *
 * \return      Uncompressed size of the Metadata in encoded form. This value
 *              may be passed to Block encoder as Uncompressed Size when using
 *              Metadata filter. On error, zero is returned.
 */
extern uint64_t lzma_metadata_size(const lzma_metadata *metadata);


extern lzma_ret lzma_metadata_decoder_init(lzma_metadata_decoder **coder,
		lzma_allocator *allocator, lzma_metadata *metadata,
		lzma_bool want_extra);

extern lzma_ret lzma_metadata_decode(
		lzma_metadata_decoder *coder, lzma_stream *strm_in);

extern void lzma_metadata_decoder_end(
		lzma_metadata_decoder **coder, lzma_allocator *allocator);


extern lzma_bool lzma_metadata_index_equal(
		const lzma_index *a, const lzma_index *b);


/**********
 * Blocks *
 **********/

/**
 * \brief       The .lzma Block options
 */
typedef struct {
	/* Encoding/Decoding options */

	/**
	 * \brief       Type of integrity Check
	 *
	 * The type of the integrity Check is not stored into the Block
	 * Header, thus its value must be provided also when decoding.
	 */
	lzma_check_type check;

	/**
	 * \brief       Precense of CRC32 of the Block Header
	 *
	 * Set this to true if CRC32 of the Block Header should be
	 * calculated and stored in the Block Header.
	 *
	 * There is no way to autodetect if CRC32 is present in the Block
	 * Header, thus this information must be provided also when decoding.
	 */
	lzma_bool has_crc32;

	/**
	 * \brief       Uncompressed Size in bytes
	 *
	 * Encoder: If this value is not UINT64_MAX, it is stored to
	 * the Uncompressed Size field in the Block Header. The real
	 * uncompressed size of the data being compressed must match
	 * the Uncompressed Size or LZMA_HEADER_ERROR is returned.
	 *
	 * If Uncompressed Size is unknown, End of Payload Marker must
	 * be used. If uncompressed_size == UINT64_MAX and has_eopm == 0,
	 * LZMA_HEADER_ERROR will be returned.
	 *
	 * Decoder: If this value is not UINT64_MAX, it is compared
	 * to the real Uncompressed Size. If they do not match,
	 * LZMA_HEADER_ERROR is returned.
	 */
	uint64_t uncompressed_size;

	/**
	 * \brief       Size of the Compressed Data in bytes
	 *
	 * This flag is useful only for advanced uses. Unless you know
	 * what you are doing, set this to UINT64_MAX.
	 *
	 * \todo        Document this so that advanced users know how
	 *              to use it. ;-)
	 */
	uint64_t compressed_size;

	/*
	 * Options needed only when encoding
	 *
	 * The .lzma Block decoder ignores these options so they can be
	 * left uninitialized when decoding.
	 */

	/**
	 * \brief       Usage of End of Payload Marker
	 *
	 * If this is true, End of Payload Marker is used.
	 */
	lzma_bool has_eopm;

	/**
	 * \brief       True if the Block is a Metadata Block
	 *
	 * If this is true, the Metadata bit will be set in the Block Header.
	 * It is up to the application to store correctly formatted data
	 * into Metadata Block.
	 */
	lzma_bool is_metadata;

	/**
	 * \brief       Alignment of the beginning of the Block
	 *
	 * See lzma_options_stream_multi.alignment for more information.
	 */
	uint32_t alignment;

	/**
	 * \brief       Array of filters
	 *
	 * There can be at maximum of seven filters. The end of the array
	 * is marked with .id = UINT64_MAX. Minimum number of filters is
	 * zero; in that case, an implicit Copy filter is used.
	 */
	lzma_options_filter filters[8];

} lzma_options_block;


/**
 * \brief       Initializes .lzma Block encoder
 *
 * This function is required for multi-thread encoding. It may also be
 * useful when implementing custom file formats.
 */
extern lzma_ret lzma_block_encoder(
		lzma_stream *strm, const lzma_options_block *options);


/**
 * \brief       Initializes decoder for .lzma Block
 *
 * \return      - LZMA_OK: Initialization was successful.
 *              - LZMA_UNSUPPORTED_CHECK: Initialization was successful, but
 *                the given Check type is not supported, thus Check will be
 *                ignored.
 *              - LZMA_STREAM_ERROR: Something in *options is outside the
 *                valid range (uncompressed_size, compressed_size or check).
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 *                Initialization failed.
 */
extern lzma_ret lzma_block_decoder(
		lzma_stream *strm, const lzma_options_block *options);


/**************
 * Raw coding *
 **************/

/**
 * \brief       Initializes raw encoder
 *
 * This function may be useful when implementing custom file formats.
 *
 * \param       strm    Pointer to properly prepared lzma_stream
 * \param       options Array of lzma_options_filter structures.
 *                      The end of the array must be marked with
 *                      .id = UINT64_MAX. The minimum number of
 *                      filters is zero; the maximum is determined
 *                      by available memory.
 *
 * \todo FIXME desc
 */
extern lzma_ret lzma_raw_encoder(
		lzma_stream *strm, const lzma_options_filter *options,
		uint64_t uncompressed_size, lzma_bool implicit);


/**
 * \brief       Initializes raw decoder
 *
 * \param       strm    Pointer to properly prepared lzma_stream
 * \param       options Array of filter options; end of array is
 *                      indicated with .id = UINT64_MAX.
 *
 * \todo FIXME desc
 */
extern lzma_ret lzma_raw_decoder(
		lzma_stream *strm, const lzma_options_filter *options,
		uint64_t uncompressed_size, lzma_bool implicit);


/****************************
 * Stream Header and Footer *
 ****************************/

/**
 * Options for encoding and decoding Stream Header and Stream Footer
 */
typedef struct {
	/* Options need by both Stream Header and Stream Footer encoders */

	/**
	 * Type of the Check calculated from uncompressed data
	 */
	lzma_check_type check;

	/**
	 * True if Block Headers have the CRC32 field. Note that the CRC32
	 * field is always present in the Stream Header.
	 */
	lzma_bool has_crc32;

	/**
	 * True if the Stream is a Multi-Block Stream.
	 */
	lzma_bool is_multi;

	/* Options need by only Stream Footer encoder */

	/**
	 * Uncompressed Size of a Single-Block Stream if Uncompressed Size
	 * isn't already stored in the Block Header. Set this to UINT64_MAX
	 * to not write Uncompressed Size field into Stream Footer of
	 * Single-Block Stream; Uncompressed Size is never written if is_multi
	 * has been set.
	 *
	 * In the decoder, this is the only field that needs to be initialized
	 * by the application before calling lzma_stream_footer_decoder().
	 * If uncompressed_size has been set to zero, it will be parsed
	 * from the input; otherwise the decoder will assume that Uncompressed
	 * Size is not present, and will leave this variable untouched.
	 */
	uint64_t uncompressed_size;

	/**
	 * Backward Size is always needed in Stream Footer. The size of the
	 * Uncompressed Size field (if present) is automatically added
	 * by the Stream Header encoder, so applications must not calculate
	 * the size of Uncompressed Size themselves.
	 */
	uint64_t backward_size;

} lzma_stream_flags;


/**
 * \brief       Allocates and encodes Stream Header
 *
 * Encoding of the Stream Header is done with a single call instead of
 * first initializing and then doing the actual work with lzma_code().
 *
 * \param       header      Pointer to a pointer that will hold the
 *                          allocated buffer. Caller must free it once
 *                          it isn't needed anymore.
 * \param       header_size Pointer to a variable that will the final size
 *                          of the header buffer.
 * \param       allocator   lzma_allocator for custom allocator functions.
 *                          Set to NULL to use malloc().
 * \param       options     Stream Header options to be encoded.
 *
 * \return      - LZMA_OK: Success; *header and *header_size set.
 *              - LZMA_HEADER_ERROR: *options is invalid.
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 */
extern lzma_ret lzma_stream_header_encode(
		uint8_t **header, size_t *header_size,
		lzma_allocator *allocator, const lzma_stream_flags *options);

/**
 * \brief       Allocates and encodes Stream Footer
 *
 * \param       footer      Pointer to a pointer that will hold the
 *                          allocated buffer. Caller must free it once
 *                          it isn't needed anymore.
 * \param       footer_size Pointer to a variable that will the final size
 *                          of the footer buffer.
 * \param       allocator   lzma_allocator for custom allocator functions.
 *                          Set to NULL to use malloc().
 * \param       options     Stream Header options to be encoded.
 *
 * \return      - LZMA_OK: Success; *header and *header_size set.
 *              - LZMA_HEADER_ERROR: *options is invalid.
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 */
extern lzma_ret lzma_stream_footer_encode(
		uint8_t **footer, size_t *footer_size,
		lzma_allocator *allocator, const lzma_stream_flags *options);


/**
 * \brief       Initializes Stream Header decoder
 *
 * \param       strm        Pointer to lzma_stream used to pass input data
 * \param       options     Target structure for parsed results
 *
 * \return      - LZMA_OK: Successfully initialized
 *              - LZMA_MEM_ERROR: Cannot allocate memory
 *
 * The actual decoding is done with lzma_code() and freed with lzma_end().
 */
extern lzma_ret lzma_stream_header_decoder(
		lzma_stream *strm, lzma_stream_flags *options);


/**
 * \brief       Initializes Stream Footer decoder
 *
 * \param       strm        Pointer to lzma_stream used to pass input data
 * \param       options     Target structure for parsed results.
 * \param       decode_uncompressed_size
 *                          Set to true if the first field to decode is
 *                          Uncompressed Size. Set to false if the first
 *                          field to decode is Backward Size.
 *
 * \return      - LZMA_OK: Successfully initialized
 *              - LZMA_MEM_ERROR: Cannot allocate memory
 *
 * The actual decoding is done with lzma_code() and freed with lzma_end().
 */
extern lzma_ret lzma_stream_footer_decoder(
		lzma_stream *strm, lzma_stream_flags *options,
		lzma_bool decode_uncompressed_size);


/**
 * \brief       Decodes Stream Footer starting at end of the given buffer
 *
 * Decoding the Stream Footer backwards is needed when reading a Stream
 * backwards for example to get information from Footer Metadata Block.
 *
 * \param       in          Beginning of the input buffer
 * \param       in_size     At maximum, bytes from in[0] to in[*in_size - 1]
 *                          will be used. On success, *in_size is set so that
 *                          in[*in_size] is the first byte of the Backward
 *                          Size field. Thus,
 *                          in[*in_size - options->backward_size] should be
 *                          the first byte of the Block.
 * \param       options     Data structure to hold the parsed results.
 *
 * This function will always _try_ to parse the Uncompressed Size
 * field. However, it doesn't exists in all Streams, and parsing
 * of that field will fail. Because this kind of fail is not fatal,
 * lzma_stream_footer_reverse_decode() will still succeed.
 * options->uncompressed_size will be set to UINT64_MAX if parsing fails.
 *
 * Application must check the Block Header before deciding if
 * options->uncompressed_size could be a valid value.
 *
 * \todo        This description could be clearer.
 */
extern lzma_ret lzma_stream_footer_reverse_decode(const uint8_t *in,
		size_t *in_size, lzma_stream_flags *options);


/****************
 * Block Header *
 ****************/

/**
 * \brief       Allocates a buffer and encodes Block Header into the buffer
 *
 * Encoding of the Block options is done with a single call instead of
 * first initializing and then doing the actual work with lzma_code().
 *
 * \param       header      Pointer to a pointer that will hold the
 *                          allocated buffer. Caller must free it once
 *                          it isn't needed anymore.
 * \param       header_size Pointer to a variable that will the final size
 *                          of the header buffer.
 * \param       allocator   lzma_allocator for custom allocator functions.
 *                          Set to NULL to use malloc().
 * \param       options     Block options to be encoded.
 *
 * \return      - LZMA_OK: Success; *header and *header_size set.
 *              - LZMA_HEADER_ERROR: Invalid or unsupported Block options
 *              - LZMA_MEM_ERROR: Cannot allocate memory.
 */
extern lzma_ret lzma_block_header_encode(uint8_t **header, size_t *header_size,
		lzma_allocator *allocator, const lzma_options_block *options);


/**
 * \brief       Initializes Block Header decoder
 *
 * The Block Header field is the in beginning of Block in the .lzma format.
 * You may use it also in your custom file formats.
 *
 * The parsed result is stored into *options. options->filters[].options
 * are initialized but old values are NOT free()d. options->check and
 * options->has_crc32 are not touched, because their values are not available
 * in the Block Header.
 *
 * Because the results of this decoder are placed into *options,
 * strm->next_in, strm->avail_in, and strm->total_in are not used.
 *
 * The only valid `action' with lzma_code() is LZMA_RUN.
 */
extern lzma_ret lzma_block_header_decoder(
		lzma_stream *strm, lzma_options_block *options);


/****************
 * Filter Flags *
 ****************/

/**
 * \brief       Calculates encoded size of a Filter Flags field
 *
 * Knowing the size of Filter Flags is useful to know when allocating
 * memory to hold the encoded Filter Flags.
 *
 * \note        If you need to calculate size of List of Filter Flags,
 *              you need to loop over every lzma_options_filter entry.
 */
extern size_t lzma_filter_flags_size(const lzma_options_filter *options);


/**
 * \brief       Encodes Filter Flags into given buffer
 *
 * In contrast to some functions, this doesn't allocate the needed buffer.
 * This is due to how this function is used internally by liblzma.
 *
 * \param       out         Beginning of the output buffer
 * \param       out_pos     out[*out_pos] is the next write position. This
 *                          is updated by the encoder.
 * \param       out_size    out[out_size] is the first byte to not write.
 * \param       options     Filter options to be encoded
 *
 * \return      - LZMA_OK: Encoding was successful.
 *              - LZMA_HEADER_ERROR: Invalid or unsupported *options.
 *              - LZMA_BUF_ERROR: Not enough output buffer space.
 */
extern lzma_ret lzma_filter_flags_encode(uint8_t *out, size_t *out_pos,
		size_t out_size, const lzma_options_filter *options);


/**
 * \brief       Initializes Filter Flags decoder
 *
 * The Filter Flags field can be found from Block Header of the .lzma format.
 * You may use it also in your custom file formats.
 *
 * The parsed result is stored into *options. options->options is initialized
 * but the old value is NOT free()d.
 *
 * Because the results of this decoder are placed into *options,
 * strm->next_in, strm->avail_in, and strm->total_in are not used.
 *
 * The only valid `action' with lzma_code() is LZMA_RUN.
 */
extern lzma_ret lzma_filter_flags_decoder(
		lzma_stream *strm, lzma_options_filter *options);



/******************************************
 * Variable length integer (VLI) handling *
 ******************************************/

/**
 * \brief       Maximum supported value of variable-length integer
 */
#define LZMA_VLI_VALUE_MAX (UINT64_MAX / 2)

/**
 * \brief       VLI value to denote that the value is unknown
 */
#define LZMA_VLI_VALUE_UNKNOWN UINT64_MAX

/**
 * \brief       Maximum supported length of variable length integers
 */
#define LZMA_VLI_BYTES_MAX 9


/**
 * \brief       Simple macro to validate variable-length integer
 *
 * This is useful to test that application has given acceptable values
 * for example in the uncompressed_size and compressed_size variables.
 *
 * \return      True if the integer is representable as VLI or if it
 *              indicates unknown value.
 */
#define lzma_vli_is_valid(vli) \
	((vli) <= LZMA_VLI_VALUE_MAX || (vli) == LZMA_VLI_VALUE_UNKNOWN)


/**
 * \brief       Encodes variable-length integer
 *
 * In the new .lzma format, most integers are encoded in variable-length
 * representation. This saves space when smaller values are more likely
 * than bigger values.
 *
 * The encoding scheme encodes seven bits to every byte, using minimum
 * number of bytes required to represent the given value. In other words,
 * it puts 7-63 bits into 1-9 bytes. This implementation limits the number
 * of bits used to 63, thus num must be at maximum of INT64_MAX (not
 * UINT64_MAX). You may use LZMA_VLI_VALUE_MAX for clarity.
 *
 * \param       vli       Integer to be encoded
 * \param       vli_pos   How many bytes have already been written out. This
 *                        must be less than 9 before calling this function.
 * \param       vli_size  Minimum size that the variable-length representation
 *                        must take. This is useful if you want to use
 *                        variable-length integers as padding. Usually you want
 *                        to set this to zero. The maximum allowed value is 9.
 * \param       out       Beginning of the output buffer
 * \param       out_pos   The next byte will be written to out[*out_pos].
 * \param       out_size  Size of the out buffer; the first byte into
 *                        which no data is written to is out[out_size].
 *
 * \return      - LZMA_OK: So far all OK, but the integer is not
 *                completely written out yet.
 *              - LZMA_STREAM_END: Integer successfully encoded.
 *              - LZMA_BUF_ERROR: No output space (*out_pos == out_size)
 *              - LZMA_STREAM_ERROR: Arguments are not sane.
 */
extern lzma_ret lzma_vli_encode(
		uint64_t vli, size_t *lzma_restrict vli_pos, size_t vli_size,
		uint8_t *lzma_restrict out, size_t *lzma_restrict out_pos,
		size_t out_size);


/**
 * \brief       Decodes variable-length integer
 *
 * \param       vli       Pointer to decoded integer. The decoder will
 *                        initialize it to zero when *vli_pos == 0, so
 *                        application isn't required to initialize *vli.
 * \param       vli_pos   How many bytes have already been decoded. When
 *                        starting to decode a new integer, *vli_pos must
 *                        be initialized to zero.
 * \param       in        Beginning of the input buffer
 * \param       in_pos    The next byte will be read from in[*in_pos].
 * \param       in_size   Size of the input buffer; the first byte that
 *                        won't be read is in[in_size].
 *
 * \return      - LZMA_OK: So far all OK, but the integer is not
 *                completely decoded yet.
 *              - LZMA_STREAM_END: Integer successfully decoded.
 *              - LZMA_BUF_ERROR: No input data (*in_pos == in_size)
 *              - LZMA_DATA_ERROR: Integer is longer than nine bytes.
 *              - LZMA_STREAM_ERROR: Arguments are not sane.
 */
extern lzma_ret lzma_vli_decode(uint64_t *lzma_restrict vli,
		size_t *lzma_restrict vli_pos, const uint8_t *lzma_restrict in,
		size_t *lzma_restrict in_pos, size_t in_size);


/**
 * \brief       Decodes variable-length integer reading buffer backwards
 *
 * The variable-length integer encoding is designed so that it can be read
 * either from the beginning to the end, or from the end to the beginning.
 * This feature is needed to make the Stream parseable backwards;
 * specifically, to read the Backward Size field in Stream Footer.
 *
 * \param       vli       Pointer to variable to hold the decoded integer.
 * \param       in        Beginning of the input buffer
 * \param       in_size   Number of bytes available in the in[] buffer.
 *                        On successful decoding, this is updated to match
 *                        the number of bytes used. (in[*in_size - 1] is the
 *                        first byte to process. After successful decoding,
 *                        in[*in_size] will point to the first byte of the
 *                        variable-length integer.)
 *
 * \return      - LZMA_OK: Decoding successful
 *              - LZMA_DATA_ERROR: No valid variable-length integer was found.
 *              - LZMA_BUF_ERROR: Not enough input; in practice, this may
 *                be an indication of broken input too.
 */
extern lzma_ret lzma_vli_reverse_decode(
		uint64_t *vli, const uint8_t *in, size_t *in_size);


/**
 * \brief       Gets the minimum number of bytes required to encode vli
 *
 * \return      Number of bytes on success (1-9). If vli isn't valid,
 *              zero is returned.
 */
extern size_t lzma_vli_size(uint64_t vli);


/**********
 * Checks *
 **********/

/**
 * \brief       Calculate CRC32
 *
 * Calculates CRC32 using the polynomial from the IEEE 802.3 standard.
 *
 * \param       buf     Pointer to the input buffer
 * \param       size    Size of the input buffer
 * \param       crc     Previously returned CRC value. This is used to
 *                      calculate the CRC of a big buffer in smaller chunks.
 *                      Set to zero when there is no previous value.
 *
 * \return      Updated CRC value, which can be passed to this function
 *              again to continue CRC calculation.
 */
extern uint32_t lzma_crc32(const uint8_t *buf, size_t size, uint32_t crc);


/**
 * \brief       Calculate CRC64
 *
 * Calculates CRC64 using the polynomial from the ECMA-182 standard.
 *
 * This function is used similarly to lzma_crc32(). See its documentation.
 */
extern uint64_t lzma_crc64(const uint8_t *buf, size_t size, uint64_t crc);


/*
 * SHA256 functions are currently not exported to public API.
 * Contact the author if you think it should be.
 */


#ifdef __cplusplus
}
#endif

#endif /* ifndef LZMA_ADV_H */
