///////////////////////////////////////////////////////////////////////////////
//
/// \file       memory_limitter.c
/// \brief      Limitting memory usage
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "common.h"


/// Rounds an unsigned integer upwards to the next multiple.
#define my_ceil(num, multiple) \
	((num) + (((multiple) - ((num) % (multiple))) % (multiple)))


/// Rounds upwards to the next multiple of 2 * sizeof(void*).
/// malloc() tends to align allocations this way.
#define malloc_ceil(num) my_ceil(num, 2 * sizeof(void *))


typedef struct lzma_memory_list_s lzma_memory_list;
struct lzma_memory_list_s {
	lzma_memory_list *next;
	void *ptr;
	size_t size;
};


struct lzma_memory_limitter_s {
	size_t used;
	size_t limit;
	lzma_memory_list *list;
};


extern LZMA_API lzma_memory_limitter *
lzma_memory_limitter_create(size_t limit)
{
	if (limit < sizeof(lzma_memory_limitter))
		return NULL;

	lzma_memory_limitter *mem = malloc(sizeof(lzma_memory_limitter));

	if (mem != NULL) {
		mem->used = sizeof(lzma_memory_limitter);
		mem->limit = limit;
		mem->list = NULL;
	}

	return mem;
}


extern LZMA_API void
lzma_memory_set_limit(lzma_memory_limitter *mem, size_t limit)
{
	mem->limit = limit;
	return;
}


extern LZMA_API size_t
lzma_memory_get_limit(lzma_memory_limitter *mem)
{
	return mem->limit;
}


extern LZMA_API size_t
lzma_memory_get_used(lzma_memory_limitter *mem)
{
	return mem->used;
}


extern LZMA_API void
lzma_memory_limitter_end(lzma_memory_limitter *mem, lzma_bool free_allocated)
{
	if (mem == NULL)
		return;

	lzma_memory_list *record = mem->list;
	while (record != NULL) {
		if (free_allocated)
			free(record->ptr);

		lzma_memory_list *tmp = record;
		record = record->next;
		free(tmp);
	}

	free(mem);

	return;
}


extern LZMA_API void *
lzma_memory_alloc(lzma_memory_limitter *mem, size_t nmemb, size_t size)
{
	// While liblzma always sets nmemb to one, do this multiplication
	// to make these functions usable e.g. with zlib and libbzip2.
	// Making sure that this doesn't overflow is up to the application.
	size *= nmemb;

	// Some malloc() implementations return NULL on malloc(0). We like
	// to get a non-NULL value.
	if (size == 0)
		size = 1;

	// Calculate how much memory we are going to allocate in reality.
	// TODO: We should add some rough estimate how much malloc() needs
	// for its internal structures.
	const size_t total_size = malloc_ceil(size)
			+ malloc_ceil(sizeof(lzma_memory_list));

	// Integer overflow protection
	if (SIZE_MAX - size <= total_size)
		return NULL;

	if (mem->limit < mem->used || mem->limit - mem->used < total_size)
		return NULL;

	lzma_memory_list *record = malloc(sizeof(lzma_memory_list));
	void *ptr = malloc(size);

	if (record == NULL || ptr == NULL) {
		free(record);
		free(ptr);
		return NULL;
	}

	// Add the new entry to the beginning of the list. This should be
	// more efficient when freeing memory, because usually it is
	// "last allocated, first freed".
	record->next = mem->list;
	record->ptr = ptr;
	record->size = total_size;

	mem->list = record;
	mem->used += total_size;

	return ptr;
}


extern LZMA_API void
lzma_memory_detach(lzma_memory_limitter *mem, void *ptr)
{
	if (ptr == NULL)
		return;

	assert(mem->list != NULL);

	lzma_memory_list *record = mem->list;
	lzma_memory_list *prev = NULL;

	while (record->ptr != ptr) {
		prev = record;
		record = record->next;
	}

	if (prev != NULL)
		prev->next = record->next;
	else
		mem->list = record->next;

	assert(mem->used >= record->size);
	mem->used -= record->size;

	free(record);

	return;
}


extern LZMA_API void
lzma_memory_free(lzma_memory_limitter *mem, void *ptr)
{
	if (ptr == NULL)
		return;

	lzma_memory_detach(mem, ptr);

	free(ptr);

	return;
}
