///////////////////////////////////////////////////////////////////////////////
//
/// \file       metadata_decoder.c
/// \brief      Decodes metadata stored in Metadata Blocks
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "metadata_decoder.h"


struct lzma_metadata_decoder_s {
	enum {
		SEQ_FLAGS,
		SEQ_HEADER_METADATA_SIZE,
		SEQ_TOTAL_SIZE,
		SEQ_UNCOMPRESSED_SIZE,
		SEQ_INDEX_SIZE_1,
		SEQ_INDEX_SIZE_2,
		SEQ_INDEX_ALLOC,
		SEQ_INDEX_TOTAL,
		SEQ_INDEX_UNCOMPRESSED,
		SEQ_EXTRA_PREPARE,
		SEQ_EXTRA_ALLOC,
		SEQ_EXTRA_ID,
		SEQ_EXTRA_SIZE,
		SEQ_EXTRA_SIZE_VALIDATE,
		SEQ_EXTRA_DATA_ALLOC,
		SEQ_EXTRA_DATA_COPY,
		SEQ_EXTRA_DUMMY_ALLOC,
		SEQ_EXTRA_DUMMY_ID,
		SEQ_EXTRA_DUMMY_SIZE,
		SEQ_EXTRA_DUMMY_COPY,
		SEQ_END,
	} sequence;

	/// Position in variable-length integers
	size_t pos;

	/// Temporary variable needed to decode variables whose type
	/// is size_t instead of uint64_t.
	uint64_t tmp;

	/// Pointer to target structure to hold the parsed results.
	lzma_metadata *metadata;

	/// Number of Records in Index
	size_t index_size;

	/// Pointer to the end of the Index Record list.
	lzma_index *index_current;

	/// Total Size calculated from the Index field.
	uint64_t total_size;

	/// Uncompressed Size calculated from the Index field.
	uint64_t uncompressed_size;

	/// True once all the input has been read.
	bool end_was_reached;

	/// True if Extra is present.
	bool has_extra;

	/// True if we have been requested to store the Extra to *metadata.
	bool want_extra;

	/// Pointer to the end of the Extra Record list.
	lzma_extra *extra_tail;

	/// Dummy Extra Record used when only verifying integrity of Extra
	/// (not storing it to RAM).
	lzma_extra extra_dummy;

	/// Total number of bytes allocated for Extra Record structures and
	/// data stored into them.
	size_t extra_size;

	/// buffer[buffer_pos] is the next byte to process.
	size_t buffer_pos;

	/// buffer[buffer_size] is the first byte to not process.
	size_t buffer_size;

	/// Temporary buffer to which encoded Metadata is read before
	/// it is parsed.
	uint8_t buffer[LZMA_BUFFER_SIZE];
};


/// Reads a variable-length integer to coder->num.
#define read_vli(num) \
do { \
	const lzma_ret ret = lzma_vli_decode( \
			&coder->num, &coder->pos, \
			coder->buffer, &coder->buffer_pos, \
			coder->buffer_size); \
	if (ret != LZMA_STREAM_END) \
		return ret; \
	\
	coder->pos = 0; \
} while (0)


static lzma_ret
process(lzma_metadata_decoder *coder, lzma_allocator *allocator)
{
	while (coder->buffer_pos < coder->buffer_size)
	switch (coder->sequence) {
	case SEQ_FLAGS:
		// Reserved bits must be unset.
		if (coder->buffer[coder->buffer_pos] & 0x70)
			return LZMA_HEADER_ERROR;

		// If Size of Header Metadata is present, prepare the
		// variable for variable-length integer decoding.
		// Otherwise set it to UINT64_MAX to indicate that
		// the field isn't present.
		if (coder->buffer[coder->buffer_pos] & 0x01)
			coder->metadata->header_metadata_size = 0;
		else
			coder->metadata->header_metadata_size = UINT64_MAX;

		if (coder->buffer[coder->buffer_pos] & 0x02)
			coder->metadata->total_size = 0;
		else
			coder->metadata->total_size = UINT64_MAX;

		if (coder->buffer[coder->buffer_pos] & 0x04)
			coder->metadata->uncompressed_size = 0;
		else
			coder->metadata->uncompressed_size = UINT64_MAX;

		if (coder->buffer[coder->buffer_pos] & 0x08)
			coder->index_size = 1;
		else
			coder->index_size = 0;

		coder->has_extra
			= (coder->buffer[coder->buffer_pos] & 0x80) != 0;

		++coder->buffer_pos;
		coder->sequence = SEQ_HEADER_METADATA_SIZE;
		break;

	case SEQ_HEADER_METADATA_SIZE:
		if (coder->metadata->header_metadata_size != UINT64_MAX) {
			read_vli(metadata->header_metadata_size);
			coder->sequence = SEQ_TOTAL_SIZE;
			break;
		}

		coder->sequence = SEQ_TOTAL_SIZE;

	// Fall through

	case SEQ_TOTAL_SIZE:
		if (coder->metadata->total_size != UINT64_MAX) {
			read_vli(metadata->total_size);
			coder->sequence = SEQ_UNCOMPRESSED_SIZE;
			break;
		}

		coder->sequence = SEQ_UNCOMPRESSED_SIZE;

	// Fall through

	case SEQ_UNCOMPRESSED_SIZE:
		if (coder->metadata->uncompressed_size != UINT64_MAX) {
			read_vli(metadata->uncompressed_size);
			coder->sequence = SEQ_INDEX_SIZE_1;
			break;
		}

		coder->sequence = SEQ_INDEX_SIZE_1;

	// Fall through

	case SEQ_INDEX_SIZE_1:
		if (coder->index_size != 0) {
			read_vli(tmp);
			coder->sequence = SEQ_INDEX_SIZE_2;
			break;
		}

		coder->sequence = SEQ_EXTRA_PREPARE;
		break;

	case SEQ_INDEX_SIZE_2:
		if (coder->tmp > (uint64_t)(lzma_index_size_max))
			return LZMA_LIMIT_REACHED;

		coder->index_size = (size_t)(coder->tmp);
		coder->tmp = 0;
		coder->sequence = SEQ_INDEX_ALLOC;

	// Fall through

	case SEQ_INDEX_ALLOC: {
		if (coder->index_size == 0) {
			// If Total Size is present, it must match the sum
			// of Total Sizes in Index.
			if (coder->metadata->total_size != UINT64_MAX
					&& coder->metadata->total_size
						!= coder->total_size)
				return LZMA_DATA_ERROR;

			coder->index_current = coder->metadata->index;
			coder->sequence = SEQ_INDEX_UNCOMPRESSED;
			break;
		}

		lzma_index *i = lzma_alloc(sizeof(lzma_index), allocator);
		if (i == NULL)
			return LZMA_MEM_ERROR;

		i->next = NULL;
		i->total_size = 0;
		i->uncompressed_size = 0;

		if (coder->metadata->index == NULL)
			coder->metadata->index = i;
		else
			coder->index_current->next = i;

		coder->index_current = i;

		--coder->index_size;
		coder->sequence = SEQ_INDEX_TOTAL;
	}

	// Fall through

	case SEQ_INDEX_TOTAL:
		read_vli(index_current->total_size);

		// Prevent integer overflow. We know that the just parsed VLI
		// can be at maximum of UINT64_MAX / 2, so this simple check
		// is enough.
		coder->total_size += coder->index_current->total_size;
		if (coder->total_size > UINT64_MAX / 2)
			return LZMA_DATA_ERROR;

		coder->sequence = SEQ_INDEX_ALLOC;
		break;

	case SEQ_INDEX_UNCOMPRESSED:
		read_vli(index_current->uncompressed_size);

		coder->uncompressed_size
				+= coder->index_current->uncompressed_size;
		if (coder->uncompressed_size > UINT64_MAX / 2)
			return LZMA_DATA_ERROR;

		coder->index_current = coder->index_current->next;
		if (coder->index_current != NULL)
			break;

		// If Uncompressed Size is present, it must match
		// the sum of Uncompressed Sizes in Index.
		if (coder->metadata->uncompressed_size != UINT64_MAX
				&& coder->metadata->uncompressed_size
					!= coder->uncompressed_size)
			return LZMA_DATA_ERROR;

		coder->sequence = SEQ_EXTRA_PREPARE;
		break;

	case SEQ_EXTRA_PREPARE:
		if (!coder->has_extra) {
			coder->sequence = SEQ_END;
			return LZMA_STREAM_END;
		}

		if (!coder->want_extra) {
			coder->extra_tail = &coder->extra_dummy;
			coder->sequence = SEQ_EXTRA_DUMMY_ALLOC;
			break;
		}

		coder->sequence = SEQ_EXTRA_ALLOC;

	// Fall through

	case SEQ_EXTRA_ALLOC: {
		if (coder->extra_size >= lzma_extra_size_max
				|| sizeof(lzma_extra) > lzma_extra_size_max
					- coder->extra_size)
			return LZMA_LIMIT_REACHED;

		lzma_extra *e = lzma_alloc(sizeof(lzma_extra), allocator);
		if (e == NULL)
			return LZMA_MEM_ERROR;

		e->next = NULL;
		e->id = 0;
		e->size = 0;
		e->data = NULL;

		if (coder->metadata->extra == NULL)
			coder->metadata->extra = e;
		else
			coder->extra_tail->next = e;

		coder->extra_tail = e;
		coder->sequence = SEQ_EXTRA_ID;
	}

	// Fall through

	case SEQ_EXTRA_ID:
	case SEQ_EXTRA_DUMMY_ID:
		read_vli(extra_tail->id);

		if (coder->extra_tail->id == 0) {
			coder->extra_tail->size = 0;
			coder->extra_tail->data = NULL;
			--coder->sequence;
		} else {
			++coder->sequence;
		}

		break;

	case SEQ_EXTRA_SIZE:
	case SEQ_EXTRA_DUMMY_SIZE:
		read_vli(tmp);
		++coder->sequence;
		break;

	case SEQ_EXTRA_SIZE_VALIDATE:
		if (coder->extra_size >= lzma_extra_size_max
				|| coder->tmp > (uint64_t)(lzma_extra_size_max
					- coder->extra_size))
			return LZMA_LIMIT_REACHED;

		coder->extra_tail->size = (size_t)(coder->tmp);
		coder->tmp = 0;
		coder->sequence = SEQ_EXTRA_DATA_ALLOC;

	// Fall through

	case SEQ_EXTRA_DATA_ALLOC: {
		uint8_t *d = lzma_alloc((size_t)(coder->extra_tail->size),
				allocator);
		if (d == NULL)
			return LZMA_MEM_ERROR;

		coder->extra_tail->data = d;
		coder->sequence = SEQ_EXTRA_DATA_COPY;
	}

	// Fall through

	case SEQ_EXTRA_DATA_COPY:
		bufcpy(coder->buffer, &coder->buffer_pos, coder->buffer_size,
				coder->extra_tail->data, &coder->pos,
				(size_t)(coder->extra_tail->size));

		if ((size_t)(coder->extra_tail->size) == coder->pos) {
			coder->pos = 0;
			coder->sequence = SEQ_EXTRA_ALLOC;
		}

		break;

	case SEQ_EXTRA_DUMMY_ALLOC:
		// Not really alloc, just initialize the dummy entry.
		coder->extra_dummy = (lzma_extra){
			.next = NULL,
			.id = 0,
			.size = 0,
			.data = NULL,
		};

		coder->sequence = SEQ_EXTRA_DUMMY_ID;
		break;

	case SEQ_EXTRA_DUMMY_COPY: {
		// Simply skip as many bytes as indicated by Extra Record Size.
		// We don't check lzma_extra_size_max because we don't
		// allocate any memory to hold the data.
		const size_t in_avail = coder->buffer_size - coder->buffer_pos;
		const size_t skip = MIN((uint64_t)(in_avail), coder->tmp);
		coder->buffer_pos += skip;
		coder->tmp -= skip;

		if (coder->tmp == 0)
			coder->sequence = SEQ_EXTRA_DUMMY_ALLOC;

		break;
	}

	case SEQ_END:
		// If we get here, there is still some input data available,
		// but according to Metadata Flags we should have parsed
		// everything already.
		return LZMA_HEADER_ERROR;

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


extern lzma_ret
lzma_metadata_decode_real(lzma_metadata_decoder *coder,
		lzma_next_coder *next, lzma_allocator *allocator,
		const uint8_t *in, size_t *in_pos, size_t in_size)
{
	while (true) {
		// Fill the buffer if it is empty.
		if (coder->buffer_pos == coder->buffer_size) {
			coder->buffer_pos = 0;
			coder->buffer_size = 0;

			const lzma_ret ret = next->code(next->coder, allocator,
					in, in_pos, in_size, coder->buffer,
					&coder->buffer_size, LZMA_BUFFER_SIZE,
					LZMA_RUN);

			switch (ret) {
			case LZMA_OK:
				// Return immediatelly if we got no new data.
				if (coder->buffer_size == 0)
					return LZMA_OK;

				break;

			case LZMA_STREAM_END:
				coder->end_was_reached = true;
				break;

			default:
				return ret;
			}
		}

		// Process coder->buffer.
		const lzma_ret ret = process(coder, allocator);
		if (ret != LZMA_OK)
			return ret;

		// On success, process() eats all the input.
		assert(coder->buffer_pos == coder->buffer_size);

		if (coder->end_was_reached) {
			// Check that the sequence is not in the
			// middle of anything.
			if (coder->index_size != 0)
				return LZMA_DATA_ERROR;

			if (coder->sequence != SEQ_INDEX_ALLOC
					&& coder->sequence != SEQ_EXTRA_PREPARE
					&& coder->sequence != SEQ_EXTRA_ALLOC
					&& coder->sequence
						!= SEQ_EXTRA_DUMMY_ALLOC
					&& coder->sequence != SEQ_END)
				return LZMA_DATA_ERROR;

			return LZMA_STREAM_END;
		}
	}
}


extern LZMA_API void
lzma_metadata_decoder_end(
		lzma_metadata_decoder **coder, lzma_allocator *allocator)
{
	lzma_free(*coder, allocator);
	*coder = NULL;
	return;
}


extern LZMA_API lzma_ret
lzma_metadata_decoder_init(lzma_metadata_decoder **coder,
		lzma_allocator *allocator, lzma_metadata *metadata,
		lzma_bool want_extra)
{
	if (metadata == NULL)
		return LZMA_STREAM_ERROR;

	*coder = lzma_alloc(sizeof(lzma_metadata_decoder), allocator);
	if (*coder == NULL)
		return LZMA_MEM_ERROR;

	(*coder)->end_was_reached = false;
	(*coder)->sequence = SEQ_FLAGS;
	(*coder)->pos = 0;
	(*coder)->tmp = 0;
	(*coder)->metadata = metadata;
	(*coder)->index_current = NULL;
	(*coder)->total_size = 0;
	(*coder)->uncompressed_size = 0;
	(*coder)->want_extra = want_extra;
	(*coder)->extra_tail = NULL;
	(*coder)->extra_size = 0;
	(*coder)->buffer_pos = 0;
	(*coder)->buffer_size = 0;

	return LZMA_OK;
}


extern LZMA_API lzma_bool
lzma_metadata_index_equal(const lzma_index *a, const lzma_index *b)
{
	while (a != NULL && b != NULL) {
		if (a->uncompressed_size != b->uncompressed_size
				|| a->total_size != b->total_size)
			return false;

		a = a->next;
		b = b->next;
	}

	if (a != NULL || b != NULL)
		return false;

	return true;
}


/*
extern LZMA_API lzma_bool
lzma_metdata_compare(const lzma_metadata *m1, const lzma_metadata *m2)
{
	if (m1->header_metadata_size != UINT64_MAX
			&& m2->header_metadata_size != UINT64_MAX
			&& m1->header_metadata_size
				!= m2->header_metadata_size)
		return false;

	uint64_t total_size = 0;
	uint64_t uncompressed_size = 0;

	{
		const lzma_index *i1 = m1->index;
	}
}
*/
