///////////////////////////////////////////////////////////////////////////////
//
/// \file       subblock_decoder.c
/// \brief      Decoder of the Subblock filter
//
//  Copyright (C) 2007 Lasse Collin
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
///////////////////////////////////////////////////////////////////////////////

#include "subblock_decoder.h"
#include "subblock_decoder_helper.h"
#include "raw_decoder.h"


struct lzma_coder_s {
	lzma_next_coder next;

	enum {
		SEQ_FLAGS,
		SEQ_SIZE_1,
		SEQ_SIZE_2,
		SEQ_SIZE_3,
		SEQ_UNCOMPRESSED_SIZE,
		SEQ_FILTER_FLAGS,
		SEQ_FILTER_INIT,
		SEQ_COPY,
	} sequence;

	/// Number of bytes left in the current Subblock Data field.
	size_t size;

	/// Uncompressed Size, or UINT64_MAX if unknown.
	uint64_t uncompressed_size;

	/// True when .next.code() has returned LZMA_STREAM_END.
	bool end_was_reached;

	/// Options for the Subblock decoder helper. This is used to tell
	/// the helper when it should return LZMA_STREAM_END to the subfilter.
	lzma_options_subblock_helper helper;

	/// True if Subfilters are allowed.
	bool allow_subfilters;

	/// Possible subfilter
	lzma_next_coder subfilter;

	/// Uncompressed Size of the subfilter.
	uint64_t subfilter_uncompressed;

	/// Position in the variable-length integer when decoding
	/// subfilter_uncompressed.
	size_t subfilter_uncompressed_pos;

	/// Filter Flags decoder is needed to parse the ID and Properties
	/// of the subfilter.
	lzma_next_coder filter_flags_decoder;

	/// The filter_flags_decoder stores its results here.
	lzma_options_filter filter_flags;

	/// buffer[buffer_pos] is the next byte to use from buffer[].
	size_t buffer_pos;

	/// buffer[buffer_size] is the first byte to not use from buffer[].
	size_t buffer_size;

	/// Temporary buffer needed when the Subblock filter is not the last
	/// filter in the chain. The output of the next filter is first
	/// decoded into buffer[], which is then used as input for the actual
	/// Subblock decoder.
	uint8_t buffer[LZMA_BUFFER_SIZE];
};


/// Values of valid Subblock Flags
enum {
	FLAG_SIZE,
	FLAG_PADDING,
	FLAG_EOPM,
	FLAG_NEW_WITH_KNOWN_UNCOMPRESSED,
	FLAG_NEW_WITH_UNKNOWN_UNCOMPRESSED,
	FLAG_END_SUBFILTER,
};


/// Calls the subfilter and updates size of Subblock Data (coder->size).
static lzma_ret
subfilter_decode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *in, size_t *in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	assert(coder->subfilter.code != NULL);

	// Store the current output position, which is needed to update
	// size of Subblock Data.
	const size_t in_start = *in_pos;

	// Call the subfilter.
	const lzma_ret ret = coder->subfilter.code(
			coder->subfilter.coder, allocator,
			in, in_pos, in_size, out, out_pos, out_size, action);

	// Update the remaining size of the Subblock Data.
	assert(*in_pos - in_start <= coder->size);
	coder->size -= *in_pos - in_start;

	return ret;
}


static lzma_ret
decode_buffer(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *in, size_t *in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	while (*out_pos < out_size && (*in_pos < in_size
			|| coder->sequence == SEQ_COPY))
	switch (coder->sequence) {
	case SEQ_FLAGS: {
		// Check that bits 0-3 for Size are unset when Subblock Type
		// is not Data.
		if ((in[*in_pos] & 0xF0) && (in[*in_pos] & 0x0F))
			return LZMA_DATA_ERROR;

		// Do the correct action depending on the Subblock Type.
		switch (in[*in_pos] >> 4) {
		case FLAG_SIZE:
			// First four bits of the Subblock Data size.
			coder->size = in[*in_pos];
			coder->sequence = SEQ_SIZE_1;
			break;

		case FLAG_PADDING:
			// Nothing to do.
			break;

		case FLAG_EOPM:
			// There must be no Subfilter enabled.
			if (coder->subfilter.code != NULL)
				return LZMA_DATA_ERROR;

			// End of Payload Marker must not be used if
			// uncompressed size is known.
			if (coder->uncompressed_size != UINT64_MAX)
				return LZMA_DATA_ERROR;

			++*in_pos;
			return LZMA_STREAM_END;

		case FLAG_NEW_WITH_KNOWN_UNCOMPRESSED: {
			if (coder->subfilter.code != NULL
					|| !coder->allow_subfilters)
				return LZMA_DATA_ERROR;

			// Initialize Filter Flags decoder.
			const lzma_ret ret = lzma_filter_flags_decoder_init(
					&coder->filter_flags_decoder,
					allocator, &coder->filter_flags);
			if (ret != LZMA_OK)
				return ret;

			// Prepare the variables for Uncompressed Size parsing.
			coder->subfilter_uncompressed = 0;
			coder->subfilter_uncompressed_pos = 0;

			coder->sequence = SEQ_UNCOMPRESSED_SIZE;
			break;
		}

		case FLAG_NEW_WITH_UNKNOWN_UNCOMPRESSED: {
			if (coder->subfilter.code != NULL
					|| !coder->allow_subfilters)
				return LZMA_DATA_ERROR;

			// Initialize Filter Flags decoder.
			const lzma_ret ret = lzma_filter_flags_decoder_init(
					&coder->filter_flags_decoder,
					allocator, &coder->filter_flags);
			if (ret != LZMA_OK)
				return ret;

			// Mark that Uncompressed Size is unknown, and prepare
			// Filter Flags decoder.
			coder->subfilter_uncompressed = UINT64_MAX;

			coder->sequence = SEQ_FILTER_FLAGS;
			break;
		}

		case FLAG_END_SUBFILTER:
			// Return an error if we don't have any subfilter in
			// use at the moment, or if Uncompressed Size is
			// known for the subfilter.
			if (coder->subfilter.code == NULL
					|| coder->subfilter_uncompressed
						!= UINT64_MAX)
				return LZMA_DATA_ERROR;

			// Tell the helper filter to indicate End of Input
			// to our subfilter.
			coder->helper.end_was_reached = true;

			const size_t out_start = *out_pos;
			size_t dummy = 0;
			const lzma_ret ret = subfilter_decode(coder, allocator,
					NULL, &dummy, 0, out, out_pos,out_size,
					action);

			// Check and update uncompressed size.
			if (coder->uncompressed_size != UINT64_MAX) {
				const size_t out_used = *out_pos - out_start;

				if ((uint64_t)(out_used)
						> coder->uncompressed_size)
					return LZMA_DATA_ERROR;

				coder->uncompressed_size -= out_used;
			}

			// If we didn't reach the end of the subfilter's output
			// yet, return to the application. On the next call we
			// will get to this same switch-case again, because we
			// haven't updated *in_pos yet.
			if (ret != LZMA_STREAM_END)
				return ret;

			// Free subfilter's memory.
			lzma_next_coder_end(&coder->subfilter, allocator);

			// Free memory used for subfilter options.
			lzma_next_coder_end(&coder->filter_flags_decoder,
					allocator);

			// Check if we have produced enough output now.
			if (coder->uncompressed_size == 0) {
				++*in_pos;
				return LZMA_STREAM_END;
			}

			break;

		default:
			return LZMA_DATA_ERROR;
		}

		++*in_pos;
		break;
	}

	case SEQ_SIZE_1:
		coder->size |= (size_t)(in[*in_pos]) << 4;
		++*in_pos;
		coder->sequence = SEQ_SIZE_2;
		break;

	case SEQ_SIZE_2:
		coder->size |= (size_t)(in[*in_pos]) << 12;
		++*in_pos;
		coder->sequence = SEQ_SIZE_3;
		break;

	case SEQ_SIZE_3:
		coder->size |= (size_t)(in[*in_pos]) << 20;
		++*in_pos;
		coder->sequence = SEQ_COPY;
		break;

	case SEQ_UNCOMPRESSED_SIZE: {
		const lzma_ret ret = lzma_vli_decode(
				&coder->subfilter_uncompressed,
				&coder->subfilter_uncompressed_pos,
				in, in_pos, in_size);
		if (ret != LZMA_STREAM_END)
			return ret;

		coder->sequence = SEQ_FILTER_FLAGS;
		break;
	}

	case SEQ_FILTER_FLAGS: {
		const lzma_ret ret = coder->filter_flags_decoder.code(
				coder->filter_flags_decoder.coder, allocator,
				in, in_pos, in_size, NULL, NULL, 0, LZMA_RUN);
		if (ret != LZMA_STREAM_END)
			return ret == LZMA_HEADER_ERROR
					? LZMA_DATA_ERROR : ret;

		// Initialize in its own sequence step to make it restartable
		// in case of LZMA_MEM_ERROR.
		coder->sequence = SEQ_FILTER_INIT;
		break;
	}

	case SEQ_FILTER_INIT: {
		// Subblock and Copy filters are not allowed.
		if (coder->filter_flags.id == LZMA_FILTER_COPY
				|| coder->filter_flags.id
					== LZMA_FILTER_SUBBLOCK)
			return LZMA_DATA_ERROR;

		coder->helper.end_was_reached = false;

		lzma_options_filter filters[3] = {
			{
				.id = coder->filter_flags.id,
				.options = coder->filter_flags.options,
			}, {
				.id = LZMA_FILTER_SUBBLOCK_HELPER,
				.options = &coder->helper,
			}, {
				.id = UINT64_MAX,
				.options = NULL,
			}
		};

		// Use the helper filter only when the Subblock filter
		// must be able to indicate End of Input to the Subfilter
		// (that is, Uncompressed Size when is is unknown).
		//
		// Optimization: We know that LZMA uses End of Payload Marker
		// (not End of Input), so we can omit the helper filter even
		// when using LZMA with unknown Uncompressed Size.
		if (coder->subfilter_uncompressed != UINT64_MAX
				|| filters[0].id == LZMA_FILTER_LZMA)
			filters[1].id = UINT64_MAX;

		const lzma_ret ret = lzma_raw_decoder_init(&coder->subfilter,
				allocator, filters,
				coder->subfilter_uncompressed, false);
		if (ret != LZMA_OK)
			return ret;

		coder->sequence = SEQ_COPY;
	}

	// Fall through

	case SEQ_COPY: {
		// Store the current output position for updating
		// uncompressed size.
		const size_t out_start = *out_pos;

		// Limit the amount of input to match the available
		// Subblock Data size.
		size_t in_limit;
		if (in_size - *in_pos > coder->size)
			in_limit = *in_pos + coder->size;
		else
			in_limit = in_size;

		if (coder->subfilter.code == NULL) {
			// No subfilter is enabled. Copy as much as possible
			// and update the number of bytes left in the current
			// Subblock Data field.
			coder->size -= bufcpy(in, in_pos, in_limit,
					out, out_pos, out_size);

		} else {
			// Decode more from in[] to out[] with the subfilter.
			const lzma_ret ret = subfilter_decode(coder, allocator,
					in, in_pos, in_limit,
					out, out_pos, out_size, action);

			// Check and update subfilter's uncompressed size.
			if (coder->subfilter_uncompressed != UINT64_MAX) {
				const size_t out_used = *out_pos - out_start;
				if ((uint64_t)(coder->subfilter_uncompressed)
						< out_used)
					return LZMA_DATA_ERROR;

				coder->subfilter_uncompressed -= out_size;
			}

			if (ret == LZMA_STREAM_END) {
				// There must be no End of Payload Marker
				// if Uncompressed Size is known.
				if (coder->subfilter_uncompressed != UINT64_MAX)
					return LZMA_DATA_ERROR;

				// End of subfilter can occur only at
				// a Subblock boundary.
				if (coder->size != 0)
					return LZMA_DATA_ERROR;

				lzma_next_coder_end(&coder->subfilter,
						allocator);
			}
		}

		// Check and update uncompressed size.
		if (coder->uncompressed_size != UINT64_MAX) {
			const size_t out_used = *out_pos - out_start;

			if ((uint64_t)(out_used) > coder->uncompressed_size)
				return LZMA_DATA_ERROR;

			coder->uncompressed_size -= out_used;
		}

		// If we couldn't process the whole Subblock Data yet, return.
		if (coder->size > 0)
			return LZMA_OK;

		// Check if we have decoded all the data.
		if (coder->uncompressed_size == 0
				&& coder->subfilter.code == NULL)
			return LZMA_STREAM_END;

		coder->sequence = SEQ_FLAGS;
		break;
	}

	default:
		return LZMA_STREAM_ERROR;
	}

	return LZMA_OK;
}


static lzma_ret
subblock_decode(lzma_coder *coder, lzma_allocator *allocator,
		const uint8_t *restrict in, size_t *restrict in_pos,
		size_t in_size, uint8_t *restrict out,
		size_t *restrict out_pos, size_t out_size, lzma_action action)
{
	if (coder->next.code == NULL)
		return decode_buffer(coder, allocator, in, in_pos, in_size,
				out, out_pos, out_size, action);

	while (*out_pos < out_size) {
		if (!coder->end_was_reached
				&& coder->buffer_pos == coder->buffer_size) {
			coder->buffer_pos = 0;
			coder->buffer_size = 0;

			const lzma_ret ret = coder->next.code(
					coder->next.coder,
					allocator, in, in_pos, in_size,
					coder->buffer, &coder->buffer_size,
					LZMA_BUFFER_SIZE, action);

			if (ret == LZMA_STREAM_END)
				coder->end_was_reached = true;
			else if (ret != LZMA_OK)
				return ret;
		}

		const lzma_ret ret = decode_buffer(coder, allocator,
				coder->buffer, &coder->buffer_pos,
				coder->buffer_size,
				out, out_pos, out_size, action);

		if (ret == LZMA_STREAM_END && !coder->end_was_reached)
			return LZMA_DATA_ERROR;

		if (ret != LZMA_OK)
			return ret;

		if (*in_pos == in_size)
			break;
	}

	return LZMA_OK;
}


static void
subblock_decoder_end(lzma_coder *coder, lzma_allocator *allocator)
{
	lzma_next_coder_end(&coder->next, allocator);
	lzma_next_coder_end(&coder->subfilter, allocator);
	lzma_next_coder_end(&coder->filter_flags_decoder, allocator);
	lzma_free(coder->filter_flags.options, allocator);
	lzma_free(coder, allocator);
	return;
}


extern lzma_ret
lzma_subblock_decoder_init(lzma_next_coder *next, lzma_allocator *allocator,
		const lzma_filter_info *filters)
{
	if (next->init != &lzma_subblock_decoder_init) {
		lzma_next_coder_end(next, allocator);

		next->coder = lzma_alloc(sizeof(lzma_coder), allocator);
		if (next->coder == NULL)
			return LZMA_MEM_ERROR;

		next->coder->next = LZMA_NEXT_CODER_INIT;
		next->coder->subfilter = LZMA_NEXT_CODER_INIT;
		next->coder->filter_flags_decoder = LZMA_NEXT_CODER_INIT;

	} else {
		lzma_next_coder_end(&next->coder->subfilter, allocator);
		lzma_free(next->coder->filter_flags.options, allocator);
	}

	next->coder->filter_flags.options = NULL;

	next->coder->sequence = SEQ_FLAGS;
	next->coder->uncompressed_size = filters[0].uncompressed_size;
	next->coder->end_was_reached = false;
	next->coder->buffer_pos = 0;
	next->coder->buffer_size = 0;

	if (filters[0].options != NULL)
		next->coder->allow_subfilters = ((lzma_options_subblock *)(
				filters[0].options))->allow_subfilters;
	else
		next->coder->allow_subfilters = false;

	{
		const lzma_ret ret = lzma_next_coder_init(&next->coder->next,
				allocator, filters + 1);
		if (ret != LZMA_OK) {
			subblock_decoder_end(next->coder, allocator);
			return ret;
		}
	}

	next->code = &subblock_decode;
	next->end = &subblock_decoder_end;

	return LZMA_OK;
}
